# This code modifies code from NumPy.

# Copyright (c) 2005-2024, NumPy Developers.
# Licensed under NumPy license
# https://numpy.org/doc/stable/license.html

# Modifications and additions to the original code:
# Copyright (c) 2025 Pine Tree Labs, LLC
# Licensed under the GNU General Public License v3.0

# As a combined work, use of this code requires compliance with the GNU GPL v3.0.
# The original license terms are included below for attribution:

# === NumPy License ===

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

#     * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.

#     * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.

#     * Neither the name of the NumPy Developers nor the names of any
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from typing import TYPE_CHECKING

import casadi as cs
import numpy as np
import numpy.exceptions as npex

from .._array_impl import (
    SymbolicArray,
    _as_casadi_array,
    _dispatch_array,
    _empty_like,
    _result_type,
    array,
    eye,
    ones,
    ones_like,
    zeros,
    zeros_like,
)
from .._type_inference import shape_inference, type_inference
from ._array_ops import (
    _cs_reshape,
    binary_op,
    normalize_axis_index,
    unary_op,
)
from ._array_ufunc import SUPPORTED_UFUNCS, _dot

if TYPE_CHECKING:
    from ...typing import ShapeLike


# Do not call directly - this should be called by NumPy's array
# dispatch mechanism when x is a SymbolicArray
def _reshape(x, shape, order="C"):
    if order not in {"C", "F"}:
        raise ValueError(f"Invalid order: {order}")

    new_sym = _cs_reshape(x._sym, shape, order)
    return SymbolicArray(new_sym, dtype=x.dtype, shape=shape)


def _ravel(x, order="C"):
    return np.reshape(x, (x.size,), order=order)


def _tile(x, reps):
    # This is the dispatch function for `np.tile`, so we can
    # assume that x is a SymbolicArray.

    # From NumPy docs:
    # If reps has length d, the result will have dimension of max(d, A.ndim).
    # If A.ndim < d, A is promoted to be d-dimensional by prepending new axes. So a
    # shape (3,) array is promoted to (1, 3) for 2-D replication, or shape (1, 1, 3)
    # for 3-D replication. If this is not the desired behavior, promote A to
    # d-dimensions manually before calling this function.

    # If A.ndim > d, reps is promoted to A.ndim by prepending 1’s to it. Thus for
    # an A of shape (2, 3, 4, 5), a reps of (2, 2) is treated as (1, 1, 2, 2).
    reps = tuple(reps) if np.iterable(reps) else (reps,)

    if len(reps) > 2:
        raise ValueError("Only 1D and 2D tiling is supported")

    if len(reps) == 0:
        return x

    if all(rep == 1 for rep in reps):
        return x

    d = len(reps)
    if d < x.ndim:
        # Prepend 1's to reps
        reps = (1,) * (x.ndim - d) + reps

    elif d > x.ndim:
        # Prepend 1's to x.shape
        x = np.reshape(x, (1,) * (d - x.ndim) + x.shape)

    ret_shape = tuple(int(x.shape[i] * reps[i]) for i in range(len(reps)))

    x_cs = _as_casadi_array(x)
    # CasADi arrays are always 2D, so we need might need to adjust the
    # reps before calling cs.repmat
    cs_reps = reps
    if len(cs_reps) < 2:
        cs_reps = (1,) * (2 - len(cs_reps)) + cs_reps

    ret_cs = cs.repmat(x_cs, *cs_reps)
    return SymbolicArray(ret_cs, dtype=x.dtype, shape=ret_shape)


def _broadcast_to(x, shape):
    shape = tuple(shape) if np.iterable(shape) else (shape,)

    if not shape and x.shape:
        raise ValueError("cannot broadcast a non-scalar to a scalar array")

    if any(size < 0 for size in shape):
        raise ValueError("all elements of broadcast shape must be non-negative")

    if len(shape) > 2:
        raise ValueError("Only 0-2D arrays are supported")

    if len(shape) < x.ndim:
        raise ValueError(
            f"input operand with shape {x.shape} has more dimensions than the "
            f"broadcast shape {shape}"
        )

    # From NumPy docs:
    # NumPy compares their shapes element-wise. It starts with the trailing
    # (i.e. rightmost) dimension and works its way left. Two dimensions are
    # compatible when
    # 1. they are equal, or
    # 2. one of them is 1.

    # If these conditions are not met, a ValueError: operands could not be
    # broadcast together exception is thrown, indicating that the arrays
    # have incompatible shapes.

    # Prepend ones to the shape of x if necessary
    x_shape = x.shape
    if len(shape) > len(x_shape):
        x_shape = (1,) * (len(shape) - len(x_shape)) + x_shape

    # Uses `tile` for broadcasting (this is not how NumPy works, but it's fine
    # with symbolic arrays).
    reps = []
    for i in range(len(shape)):
        if x_shape[i] == shape[i]:
            reps.append(1)
        elif x_shape[i] == 1:
            reps.append(shape[i])  # Repeat to match the target shape
        else:
            raise ValueError(f"Cannot broadcast {x_shape} to {shape}")
    return _tile(x, tuple(reps))


def _transpose(x):
    dtype = x.dtype
    shape = shape_inference("transpose", x)
    return SymbolicArray(cs.transpose(x._sym), dtype=dtype, shape=shape)


def _mmax(x, axis=None):
    if axis is None:
        # Max over all elements
        return unary_op(cs.mmax, shape_inference="unary_to_scalar")(x)
    else:
        raise NotImplementedError("max with axis not yet supported")


def _mmin(x, axis=None):
    if axis is None:
        # Min over all elements
        return unary_op(cs.mmin, shape_inference="unary_to_scalar")(x)
    else:
        raise NotImplementedError("min with axis not yet supported")


def _diag(x):
    if len(x.shape) == 1:
        # Vector to diagonal matrix
        return SymbolicArray(cs.diag(x._sym), dtype=x.dtype, shape=(len(x), len(x)))
    else:
        # Matrix to vector
        return SymbolicArray(cs.diag(x._sym), dtype=x.dtype, shape=(len(x),))


def _append(arr, values, axis=None):
    # If axis is None, both `arr` and `values` are flattened
    if axis not in {None, 0, 1}:
        raise ValueError("Only 2D arrays are supported")

    arr, values = map(array, (arr, values))
    shape = None

    if axis is None:
        arr = arr.flatten()
        values = values.flatten()
        axis = 0
        shape = (len(arr) + len(values),)

    arr_ = arr if not isinstance(arr, SymbolicArray) else arr._sym
    values_ = values if not isinstance(values, SymbolicArray) else values._sym
    dtype = _result_type(arr, values)

    _cs_append = cs.vertcat if axis == 0 else cs.horzcat
    return SymbolicArray(_cs_append(arr_, values_), dtype=dtype, shape=shape)


def _astype(arr, dtype):
    return SymbolicArray(arr._sym, dtype=dtype, shape=arr.shape)


def _atleast_1d(*arrays):
    # This function only does anything if any of the arrays are 0D (scalars)
    # The underlying CasADi arrays always have 2D shape, so we just need
    # to superficially change the SymbolicArray shape.

    if len(arrays) == 1:
        arr = arrays[0]
        if arr.ndim > 0:
            return arr
        return SymbolicArray(arr._sym, dtype=arr.dtype, shape=(1,))

    # Call NumPy for the individual arrays, since some may be numeric
    # and not SymbolicArray types
    return tuple(np.atleast_1d(arr) for arr in arrays)


def _atleast_2d(*arrays):
    # This function will do nothing unless any of the arrays are 0D or 1D
    # in which case it will add a leading dimension(s) of size 1.
    # The underlying CasADi arrays always have 2D shape, so we just need
    # to superficially change the SymbolicArray shape.

    if len(arrays) == 1:
        arr = arrays[0]
        if arr.ndim > 1:
            return arr
        if arr.ndim == 0:
            shape = (1, 1)
        elif arr.ndim == 1:
            shape = (1, arr.size)
        return SymbolicArray(arr._sym, dtype=arr.dtype, shape=shape)

    # Call NumPy for the individual arrays, since some may be numeric
    # and not SymbolicArray types
    return tuple(np.atleast_2d(arr) for arr in arrays)


def _concatenate(arrs, axis=0, dtype=None):
    if dtype is not None:
        raise NotImplementedError("dtype argument not yet supported for concatenate")
    dtype = _result_type(*arrs)

    # All arrays must have the same number of dimensions
    ndim = arrs[0].ndim
    if ndim == 0:
        raise ValueError("Zero-dimensional arrays cannot be concatenated")
    if not all(arr.ndim == ndim for arr in arrs):
        # Find the differing array
        for i, arr in enumerate(arrs):
            if arr.ndim != ndim:
                raise ValueError(
                    "All arrays must have the same number of dimensions, but "
                    f"the array at index 0 has {ndim} dimensions and the array "
                    f"at index {i} has {arr.ndim} dimensions."
                )

    # Check that the axis is in bounds
    arrs = np.atleast_1d(*arrs)
    if not isinstance(arrs, tuple):
        arrs = (arrs,)
    axis = normalize_axis_index(axis, ndim, "axis")

    # The arrays must have the same shape except for the `axis` dimension
    arr0_shape = arrs[0].shape
    match_axes = [i for i in range(ndim) if i != axis]
    for j, arr in enumerate(arrs[1:]):
        if not all(arr.shape[i] == arr0_shape[i] for i in match_axes):
            # Find the differing array
            for i in match_axes:
                if arr.shape[i] != arr0_shape[i]:
                    raise ValueError(
                        "All the input array dimensions except for the concatenation "
                        f"axis must match exactly, but along dimension {i} the array "
                        f"at index 0 has size {arr0_shape[i]} and the array at index "
                        f"{j} has size {arr.shape[i]}."
                    )

    args = [_as_casadi_array(arr) for arr in arrs]
    shape: ShapeLike = ()

    # If all arrays are 1D, then use cs.vcat
    if ndim == 1:
        shape = (sum([arr.shape[axis] for arr in arrs]),)
        result = cs.vcat(args)

    # If all arrays are 2D, then use cs.hcat or cs.vcat, depending on the axis
    elif axis == 0:
        shape = (sum([arr.shape[0] for arr in arrs]), arrs[0].shape[1])
        result = cs.vcat(args)

    elif axis == 1:
        shape = (arrs[0].shape[0], sum([arr.shape[1] for arr in arrs]))
        result = cs.hcat(args)

    return SymbolicArray(result, dtype=dtype, shape=shape)


def _hstack(arrs, dtype=None):
    # From NumPy docs:
    # "This is equivalent to concatenation along the second axis, except for 1-D
    # arrays where it concatenates along the first axis.."
    # https://github.com/numpy/numpy/blob/v2.1.0/numpy/_core/shape_base.py#L294-L363
    arrs = np.atleast_1d(*arrs)
    # As a special case, dimension 0 of 1-dimensional arrays is "horizontal"
    if arrs and arrs[0].ndim == 1:
        return np.concatenate(arrs, 0, dtype=dtype)
    else:
        return np.concatenate(arrs, 1, dtype=dtype)


def _vstack(arrs, dtype=None):
    # From NumPy docs:
    # "This is equivalent to concatenation along the first axis after 1-D arrays of
    # shape (N,) have been reshaped to (1,N)."
    # https://github.com/numpy/numpy/blob/v2.1.0/numpy/_core/shape_base.py#L287C1-L290C66
    arrs = np.atleast_2d(*arrs)
    return np.concatenate(arrs, 0, dtype=dtype)


def _stack(arrays, axis=0, dtype=None):
    # From NumPy docs:
    # Join a sequence of arrays along a new axis.
    # The axis parameter specifies the index of the new axis in the dimensions of the
    # result. For example, if axis=0 it will be the first dimension and if axis=-1 it
    # will be the last dimension.
    # https://github.com/numpy/numpy/blob/v2.1.0/numpy/_core/shape_base.py#L377-L465

    arrays = [array(arr) for arr in arrays]
    shapes = {arr.shape for arr in arrays}
    if len(shapes) != 1:
        raise ValueError("All input arrays must have the same shape")

    result_ndim = arrays[0].ndim + 1
    axis = normalize_axis_index(axis, result_ndim, "axis")
    sl = (slice(None),) * axis + (None,)
    expanded_arrays = [arr[sl] for arr in arrays]
    return np.concatenate(expanded_arrays, axis=axis, dtype=dtype)


def _array_split(arr, indices_or_sections, axis=0):
    arr = array(arr)
    dtype = arr.dtype

    try:
        len(indices_or_sections)
        indices = indices_or_sections
    except TypeError:
        sections = indices_or_sections
        # Split into equal sections
        section_length = int(arr.shape[axis] // sections)
        indices = section_length * np.arange(1, sections)

    ndim = arr.ndim
    axis = normalize_axis_index(axis, ndim, "axis")

    # Casadi *split requires additional sections compared to numpy:
    # From CasADi docs:
    # To split up an expression horizontally into n smaller expressions,
    # you need to provide, in addition to the expression being split, a
    # vector offset of length n+1. The first element of the offset vector
    # must be 0 and the last element must be the number of columns.
    # Remaining elements must follow in a non-decreasing order.
    indices = [0, *map(int, indices), arr.shape[axis]]

    # Check non-decreasing indices
    ax_sizes = []
    for ix1, ix2 in zip(indices[:-1], indices[1:]):
        if ix1 > ix2:
            raise IndexError("indices to split must be non-decreasing")
        ax_sizes.append(ix2 - ix1)

    arr = _as_casadi_array(arr)
    shapes: list[ShapeLike] = []

    if ndim == 1:
        shapes = [(s,) for s in ax_sizes]
        results = cs.vertsplit(arr, indices)

    else:
        if axis == 0:
            ncol = arr.shape[1]
            shapes = [(s, ncol) for s in ax_sizes]
            results = cs.vertsplit(arr, indices)

        elif axis == 1:
            nrow = arr.shape[0]
            shapes = [(nrow, s) for s in ax_sizes]
            results = cs.horzsplit(arr, indices)

    return tuple(
        SymbolicArray(r, dtype=dtype, shape=s) for (r, s) in zip(results, shapes)
    )


def _vsplit(arr, indices_or_sections):
    return _split(arr, indices_or_sections, 0)


def _hsplit(arr, indices_or_sections):
    return _split(arr, indices_or_sections, 1)


# https://github.com/numpy/numpy/blob/v2.2.0/numpy/lib/_shape_base_impl.py#L803-L879
def _split(ary, indices_or_sections, axis=0):
    try:
        len(indices_or_sections)
    except TypeError:
        sections = indices_or_sections
        if ary.shape[axis] % sections != 0:
            raise ValueError(
                "array split does not result in an equal division"
            ) from None
    return np.array_split(ary, indices_or_sections, axis)


def _squeeze(a, axis=None):
    # From NumPy docs:
    # Remove axes of length one from a.
    # https://github.com/numpy/numpy/blob/v2.0.0/numpy/_core/fromnumeric.py#L1564-L1632

    if axis is not None:
        raise NotImplementedError("axis argument is not yet supported")

    shape = tuple(n for n in a.shape if n > 1)
    return np.reshape(a, shape)


def _where(condition, x=None, y=None):
    if x is None or y is None:
        raise ValueError("Calling where with only one argument is not supported")

    condition, x, y = map(array, (condition, x, y))

    shape = shape_inference("broadcast", condition, x, y)
    dtype = _result_type(condition, x, y)

    # CasADi will broadcast scalars, otherwise we need to handle broadcasting
    x = np.broadcast_to(x, shape)
    y = np.broadcast_to(y, shape)
    condition = np.broadcast_to(condition, shape)

    x = x if np.prod(x.shape) <= 1 else np.reshape(x, shape)
    y = y if np.prod(y.shape) <= 1 else np.reshape(y, shape)

    # Convert to NumPy or CasADi arrays only
    condition, x, y = map(_as_casadi_array, (condition, x, y))
    result = cs.if_else(condition, x, y)
    return SymbolicArray(result, dtype=dtype, shape=shape)


def _norm(x, ord=None, axis=None, keepdims=False):
    if axis is not None:
        raise NotImplementedError("axis argument not yet supported for norm")
    if keepdims:
        raise NotImplementedError("keepdims argument not yet supported for norm")

    if ord not in {None, 1, 2, np.inf, "fro"}:
        raise ValueError("Invalid norm order")

    # CasADi will throw an error for 2-norm applied to a matrix,
    # whereas NumPy will assume the Frobenius norm if axis is None.
    # Since we don't support axis, we will just use the Frobenius norm
    if len(x.shape) == 2:
        if ord in {None, 2}:
            ord = "fro"
    else:
        if ord == "fro":
            raise ValueError("Frobenius norm only defined for matrices")
        if ord is None:
            ord = 2

    norm = {
        1: cs.norm_1,
        2: cs.norm_2,
        np.inf: cs.norm_inf,
        "fro": cs.norm_fro,
    }[ord]

    return SymbolicArray(norm(x._sym), dtype=float, shape=())


def _outer(a, b):
    """Compute the outer product of two vectors"""
    a = np.atleast_2d(a.flatten())  # shape: (1, n)
    b = np.atleast_2d(b.flatten())  # shape: (m, 1)
    return a.T @ b


# TODO:
# - use cs.interp1d?
def _interp1d(x, xp, fp, left=None, right=None, period=None, method="linear"):
    # One-dimensional interpolation.  Should be called with
    # np.interp rather than directly.

    # Convert inputs to arrays (either NumPy or SymbolicArray)
    x, xp, fp = map(array, (x, xp, fp))

    if isinstance(x, SymbolicArray) and not isinstance(x._sym, cs.MX):
        raise ValueError("SymbolicArray must use 'kind=\"MX\"' to use interp function")

    # Check for invalid input
    if xp.ndim != 1:
        raise ValueError(f"xp must be 1-dimensional but has shape {xp.shape}")

    if fp.ndim != 1:
        raise ValueError(f"fp must be 1-dimensional but has shape {fp.shape}")

    if left is None:
        left = fp[0]

    if right is None:
        right = fp[-1]

    if period is not None:
        raise NotImplementedError("period argument not yet implemented")

    # CasAdi does not suppoert interpolating symbolic data; only numeric
    if not isinstance(xp, np.ndarray) or not isinstance(fp, np.ndarray):
        raise ValueError(
            f"xp and fp must be NumPy arrays, but have type {type(xp)} and {type(fp)}"
        )

    if not np.isscalar(left) or not np.isscalar(right):
        raise ValueError(
            f"left and right must be scalars, but have type {type(left)} and "
            f"{type(right)}"
        )

    # Output shape is the same as the input shape, dtype is the promotion of
    # the input dtypes
    shape = x.shape
    dtype = type_inference("default", x, xp, fp)

    x_cs = _as_casadi_array(x)  # Map input to either NumPy or CasADi

    # Create the CasADi interpolant object
    lut = cs.interpolant("interpolate", method, [xp], fp)
    f = SymbolicArray(lut(x_cs), shape=shape, dtype=dtype)

    # Limit the ends to the left/right values
    f = np.where(x < xp[0], left, f)  # type: ignore
    f = np.where(x > xp[-1], right, f)  # type: ignore

    return f


def _sum(x, axis=None, dtype=None):
    if dtype is not None:
        raise NotImplementedError("dtype argument not yet supported for sum")

    if np.isscalar(x) or len(x.shape) == 0:
        return x

    if axis is not None and axis >= len(x.shape):
        raise npex.AxisError(
            f"axis {axis} is out of bounds for array of dimension {len(x.shape)}"
        )

    dtype = x.dtype
    shape: tuple[int, ...] = ()
    arg = x._sym if isinstance(x, SymbolicArray) else x

    if axis is None or len(x.shape) == 1:
        shape = ()
        arg = cs.reshape(arg, (-1, 1))
        res = cs.sum1(arg)

    elif axis == 0:
        shape = (x.shape[1],)
        res = cs.sum1(arg)

    elif axis == 1:
        shape = (x.shape[0],)
        res = cs.sum2(arg)

    return SymbolicArray(res, dtype=dtype, shape=shape)


# https://github.com/numpy/numpy/blob/v2.1.0/numpy/_core/numeric.py#L1528-L1747
def _cross(a, b, axisa=-1, axisb=-1, axisc=-1, axis=None):
    if axis is not None:
        axisa, axisb, axisc = (axis,) * 3

    axisa = normalize_axis_index(axisa, a.ndim, "axisa")
    axisb = normalize_axis_index(axisb, b.ndim, "axisb")
    axisc = normalize_axis_index(axisc, a.ndim, "axisc")

    if a.shape[axisa] != 3 or b.shape[axisb] != 3:
        raise ValueError("Both arrays must have shape 3 along the axis of cross")

    # CasADi does not support individual axis specification, but only the
    # `dim` argument as in MATLAB
    # (https://www.mathworks.com/help/matlab/ref/cross.html)
    if axisa != axisb:
        raise NotImplementedError(
            "cross product along different axes not yet supported"
        )

    arg1 = a._sym if isinstance(a, SymbolicArray) else a
    arg2 = b._sym if isinstance(b, SymbolicArray) else b

    dtype = type_inference("default", a, b)
    shape = shape_inference("broadcast", a, b)

    res = cs.cross(arg1, arg2, axisa + 1)

    if axisc != axisa:
        res = res.T
        shape = shape[::-1]

    return SymbolicArray(res, dtype=dtype, shape=shape)


def _swapaxes(a, axis1, axis2):
    axis1 = normalize_axis_index(axis1, a.ndim, "axis1")
    axis2 = normalize_axis_index(axis2, a.ndim, "axis2")
    if axis1 == axis2:
        return a

    # Since we're only supporting up to 2D arrays, the only possibilities
    # are that axis1 == axis2 or that we need to transpose the array
    return a.T


def _clip(a, a_min=None, a_max=None):
    if a_min is not None:
        a = np.fmax(a, a_min)
    if a_max is not None:
        a = np.fmin(a, a_max)
    return a


def _roll(a, shift, axis=None):
    a = array(a)
    axis = normalize_axis_index(axis, a.ndim, "axis") if axis is not None else axis

    if a.shape == ():
        return a

    if axis is None:
        a = a.flatten()
        axis = 0

    n = a.shape[axis]
    shift = shift % n  # Handle large or negative shifts
    if shift == 0:
        return a

    if axis == 0:
        top = a[-shift:]
        bottom = a[:-shift]
        return np.concatenate((top, bottom), axis=0)

    elif axis == 1:
        left = a[:, -shift:]
        right = a[:, :-shift]
        return np.concatenate((left, right), axis=1)


# List from numpy.testing.overrides.get_overridable_numpy_array_functions()
SUPPORTED_FUNCTIONS = {
    "array": _dispatch_array,
    "may_share_memory": NotImplemented,
    "is_busday": NotImplemented,
    "identity": NotImplemented,
    "corrcoef": NotImplemented,
    "title": NotImplemented,
    "ptp": NotImplemented,
    "ix_": NotImplemented,
    "allclose": NotImplemented,
    "translate": NotImplemented,
    "busday_offset": NotImplemented,
    "array2string": NotImplemented,
    "upper": NotImplemented,
    "i0": NotImplemented,
    "max": _mmax,
    "isclose": NotImplemented,
    "busday_count": NotImplemented,
    "zfill": NotImplemented,
    "sinc": NotImplemented,
    "rec_join": NotImplemented,
    "multi_dot": NotImplemented,
    "array_equal": NotImplemented,
    "atleast_1d": _atleast_1d,
    "msort": NotImplemented,
    "amax": _mmax,
    "datetime_as_string": NotImplemented,
    "isnumeric": NotImplemented,
    "atleast_2d": _atleast_2d,
    "array_equiv": NotImplemented,
    "median": NotImplemented,
    "isdecimal": NotImplemented,
    "equal": NotImplemented,
    "min": _mmin,
    "choose": NotImplemented,
    "percentile": NotImplemented,
    "not_equal": NotImplemented,
    "fromfunction": NotImplemented,
    "poly": NotImplemented,
    "genfromtxt": NotImplemented,
    "amin": _mmin,
    "quantile": NotImplemented,
    "greater_equal": NotImplemented,
    "roots": NotImplemented,
    "trapz": NotImplemented,
    "prod": NotImplemented,
    "asfarray": NotImplemented,
    "less_equal": NotImplemented,
    "polyint": NotImplemented,
    "cumprod": NotImplemented,
    "polyder": NotImplemented,
    "greater": NotImplemented,
    "meshgrid": NotImplemented,
    "unique": NotImplemented,
    "delete": NotImplemented,
    "intersect1d": NotImplemented,
    "array_repr": NotImplemented,
    "setxor1d": NotImplemented,
    "in1d": NotImplemented,
    "isin": NotImplemented,
    "less": NotImplemented,
    "insert": NotImplemented,
    "polyfit": NotImplemented,
    "ndim": NotImplemented,
    "fix": NotImplemented,
    "polyval": NotImplemented,
    "union1d": NotImplemented,
    "append": _append,
    "repeat": NotImplemented,
    "str_len": NotImplemented,
    "setdiff1d": NotImplemented,
    "size": lambda x: np.prod(x.shape),
    "array_str": NotImplemented,
    "polyadd": NotImplemented,
    "digitize": NotImplemented,
    "isposinf": NotImplemented,
    "add": SUPPORTED_UFUNCS["add"],
    "polysub": NotImplemented,
    "put": NotImplemented,
    "swapaxes": _swapaxes,
    "multiply": binary_op(cs.times),
    "ediff1d": NotImplemented,
    "round": NotImplemented,
    "polymul": NotImplemented,
    "isneginf": NotImplemented,
    "take_along_axis": NotImplemented,
    "mod": SUPPORTED_UFUNCS["remainder"],
    "real": NotImplemented,
    "polydiv": NotImplemented,
    "around": NotImplemented,
    "put_along_axis": NotImplemented,
    "capitalize": NotImplemented,
    "einsum_path": NotImplemented,
    "save": NotImplemented,
    "require": NotImplemented,
    "fft": NotImplemented,
    "transpose": _transpose,
    "imag": NotImplemented,
    "center": NotImplemented,
    "apply_along_axis": NotImplemented,
    "mean": NotImplemented,
    "einsum": NotImplemented,
    "apply_over_axes": NotImplemented,
    "savez": NotImplemented,
    "count": NotImplemented,
    "partition": NotImplemented,
    "iscomplex": NotImplemented,
    "sliding_window_view": NotImplemented,
    "fill_diagonal": NotImplemented,
    "expand_dims": NotImplemented,
    "savez_compressed": NotImplemented,
    "decode": NotImplemented,
    "std": NotImplemented,
    "diag_indices_from": NotImplemented,
    "column_stack": NotImplemented,
    "argpartition": NotImplemented,
    "isreal": NotImplemented,
    "loadtxt": NotImplemented,
    "encode": NotImplemented,
    "broadcast_to": _broadcast_to,
    "dstack": NotImplemented,
    "ifft": NotImplemented,
    "endswith": NotImplemented,
    "var": NotImplemented,
    "array_split": _array_split,
    "sort": NotImplemented,
    "iscomplexobj": NotImplemented,
    "savetxt": NotImplemented,
    "rfft": NotImplemented,
    "broadcast_arrays": NotImplemented,
    "expandtabs": NotImplemented,
    "round_": NotImplemented,
    "argsort": NotImplemented,
    "irfft": NotImplemented,
    "isrealobj": NotImplemented,
    "find": NotImplemented,
    "product": NotImplemented,
    "fliplr": NotImplemented,
    "hsplit": _hsplit,
    "index": NotImplemented,
    "hfft": NotImplemented,
    "vsplit": _vsplit,
    "flipud": NotImplemented,
    "argmax": NotImplemented,
    "cumproduct": NotImplemented,
    "nan_to_num": NotImplemented,
    "inner": NotImplemented,
    "isalnum": NotImplemented,
    "eye": eye,
    "dsplit": NotImplemented,
    "ihfft": NotImplemented,
    "sometrue": NotImplemented,
    "real_if_close": NotImplemented,
    "argmin": NotImplemented,
    "isalpha": NotImplemented,
    "kron": NotImplemented,
    "fftn": NotImplemented,
    "diag": _diag,
    "common_type": NotImplemented,
    "alltrue": NotImplemented,
    "isdigit": NotImplemented,
    "tile": _tile,
    "ifftn": NotImplemented,
    "diagflat": NotImplemented,
    "merge_arrays": NotImplemented,
    "searchsorted": NotImplemented,
    "log": NotImplemented,
    "islower": NotImplemented,
    "fft2": NotImplemented,
    "atleast_3d": NotImplemented,
    "histogram_bin_edges": NotImplemented,
    "tri": NotImplemented,
    "ifft2": NotImplemented,
    "isspace": NotImplemented,
    "log10": NotImplemented,
    "drop_fields": NotImplemented,
    "histogram": NotImplemented,
    "vstack": _vstack,
    "resize": NotImplemented,
    "rfftn": NotImplemented,
    "tril": NotImplemented,
    "istitle": NotImplemented,
    "logn": NotImplemented,
    "tensorsolve": NotImplemented,
    "histogramdd": NotImplemented,
    "where": _where,
    "solve": binary_op(cs.solve, shape_inference="solve"),
    "rec_drop_fields": NotImplemented,
    "rfft2": NotImplemented,
    "hstack": _hstack,
    "tensorinv": NotImplemented,
    "concatenate": _concatenate,
    "squeeze": _squeeze,
    "empty_like": _empty_like,
    "isupper": NotImplemented,
    "log2": NotImplemented,
    "triu": NotImplemented,
    "irfftn": NotImplemented,
    "inv": unary_op(cs.inv),
    "lexsort": NotImplemented,
    "rename_fields": NotImplemented,
    "zeros_like": zeros_like,
    "zeros": zeros,
    "join": NotImplemented,
    "stack": _stack,
    "diagonal": NotImplemented,
    "nanmin": NotImplemented,
    "power": NotImplemented,
    "vander": NotImplemented,
    "rot90": NotImplemented,
    "irfft2": NotImplemented,
    "matrix_power": NotImplemented,
    "take": NotImplemented,
    "ones": ones,
    "recursive_fill_fields": NotImplemented,
    "can_cast": NotImplemented,
    "append_fields": NotImplemented,
    "ljust": NotImplemented,
    "block": NotImplemented,
    "flip": NotImplemented,
    "arccos": NotImplemented,
    "histogram2d": NotImplemented,
    "trace": unary_op(cs.trace, shape_inference="unary_to_scalar"),
    "nanmax": NotImplemented,
    "cholesky": unary_op(cs.chol),
    "min_scalar_type": NotImplemented,
    "rec_append_fields": NotImplemented,
    "lower": NotImplemented,
    "average": NotImplemented,
    "arcsin": NotImplemented,
    "nanargmin": NotImplemented,
    "ones_like": ones_like,
    "qr": NotImplemented,
    "ravel": _ravel,
    "result_type": _result_type,
    "full": NotImplemented,
    "tril_indices_from": NotImplemented,
    "repack_fields": NotImplemented,
    "lstrip": NotImplemented,
    "piecewise": NotImplemented,
    "arctanh": NotImplemented,
    "eigvals": NotImplemented,
    "nanargmax": NotImplemented,
    "select": NotImplemented,
    "dot": _dot,
    "triu_indices_from": NotImplemented,
    "nonzero": NotImplemented,
    "full_like": NotImplemented,
    "structured_to_unstructured": NotImplemented,
    "eigvalsh": NotImplemented,
    "replace": NotImplemented,
    "copy": NotImplemented,
    "nansum": NotImplemented,
    "fftshift": NotImplemented,
    "vdot": NotImplemented,
    "count_nonzero": NotImplemented,
    "rfind": NotImplemented,
    "gradient": NotImplemented,
    "shape": lambda x: x.shape,
    "eig": NotImplemented,
    "unstructured_to_structured": NotImplemented,
    "nanprod": NotImplemented,
    "ifftshift": NotImplemented,
    "bincount": NotImplemented,
    "nancumsum": NotImplemented,
    "eigh": NotImplemented,
    "rindex": NotImplemented,
    "diff": NotImplemented,
    "argwhere": NotImplemented,
    "compress": NotImplemented,
    "apply_along_fields": NotImplemented,
    "interp": _interp1d,
    "ravel_multi_index": NotImplemented,
    "flatnonzero": NotImplemented,
    "svd": NotImplemented,
    "nancumprod": NotImplemented,
    "rjust": NotImplemented,
    "unravel_index": NotImplemented,
    "pad": NotImplemented,
    "cond": NotImplemented,
    "nanmean": NotImplemented,
    "correlate": NotImplemented,
    "rpartition": NotImplemented,
    "linspace": NotImplemented,
    "angle": NotImplemented,
    "clip": _clip,
    "assign_fields_by_name": NotImplemented,
    "copyto": NotImplemented,
    "convolve": NotImplemented,
    "unwrap": NotImplemented,
    "matrix_rank": NotImplemented,
    "rsplit": NotImplemented,
    "nanmedian": NotImplemented,
    "outer": _outer,
    "logspace": NotImplemented,
    "putmask": NotImplemented,
    "require_fields": NotImplemented,
    "astype": _astype,
    "reshape": _reshape,
    "sum": _sum,
    "sort_complex": NotImplemented,
    "pinv": unary_op(cs.pinv, shape_inference="transpose"),
    "rstrip": NotImplemented,
    "nanpercentile": NotImplemented,
    "tensordot": NotImplemented,
    "geomspace": NotImplemented,
    "trim_zeros": NotImplemented,
    "any": NotImplemented,
    "split": _split,
    "packbits": NotImplemented,
    "slogdet": NotImplemented,
    "stack_arrays": NotImplemented,
    "roll": _roll,
    "nanquantile": NotImplemented,
    "det": unary_op(cs.det, shape_inference="unary_to_scalar"),
    "splitlines": NotImplemented,
    "extract": NotImplemented,
    "rollaxis": NotImplemented,
    "unpackbits": NotImplemented,
    "nanvar": NotImplemented,
    "find_duplicates": NotImplemented,
    "all": NotImplemented,
    "place": NotImplemented,
    "moveaxis": NotImplemented,
    "lstsq": NotImplemented,
    "startswith": NotImplemented,
    "sqrt": NotImplemented,
    "shares_memory": NotImplemented,
    "norm": _norm,
    "join_by": NotImplemented,
    "strip": NotImplemented,
    "nanstd": NotImplemented,
    "cross": _cross,
    "cov": NotImplemented,
    "cumsum": NotImplemented,
    "swapcase": NotImplemented,
}



"""Autodiff transformations"""

from typing import Callable, Sequence

from . import SymbolicArray
from ._function import FunctionCache


def grad(
    func: Callable,
    argnums: int | Sequence[int] = 0,
    name: str | None = None,
    static_argnums: int | Sequence[int] | None = None,
    static_argnames: str | Sequence[str] | None = None,
) -> Callable:
    """Create a function that evaluates the gradient of ``func``.

    Transforms a scalar-valued function into a new function that computes
    the gradient (vector of partial derivatives) with respect to one or more
    of its input arguments.

    Parameters
    ----------
    func : callable
        The function to differentiate. Should be a scalar-valued function
        (returns a single value with shape ``()``). If not already a compiled
        function, it will be compiled with specified static arguments.
    argnums : int or tuple of ints, optional
        Specifies which positional argument(s) to differentiate with respect to.
        Default is 0, meaning the first argument.
    name : str, optional
        Name for the created gradient function. If ``None``, a name is automatically
        generated based on the primal function's name.
    static_argnums : tuple of int, optional
        Specifies which positional arguments should be treated as static (not
        differentiated or traced symbolically). Only used if ``func`` is not already
        a compiled function.
    static_argnames : tuple of str, optional
        Specifies which keyword arguments should be treated as static. Only used
        if ``func`` is not already a compiled function.

    Returns
    -------
    callable
        A function that computes the gradient of ``func`` with respect to the
        specified arguments. If multiple arguments are specified in ``argnums``,
        the function returns a tuple of gradients, one for each specified argument.

    Notes
    -----
    When to use this function:
    - When you need to compute derivatives of a scalar-valued cost or objective function
    - For sensitivity analysis of scalar model outputs with respect to parameters

    Internally, CasADi chooses between forward and reverse mode automatic
    differentiation using a heuristic based on the number of required derivative
    calculations in either case.  For gradient calculations with a scalar output,
    typically reverse mode will be used.

    Conceptual model:
    This function uses automatic differentiation (AD) to efficiently compute exact
    derivatives, avoiding the numerical errors and computational cost of finite
    differencing approaches. Unlike numerical differentiation, automatic
    differentiation computes exact derivatives (to machine precision) by applying the
    chain rule to the computational graph generated from your function.

    The ``grad`` function specifically handles scalar-valued functions, returning the
    gradient as a column vector with the same shape as the input. For vector-valued
    functions, use ``jac`` (Jacobian) instead.

    Edge cases:

    - Raises ValueError if ``argnums`` contains a static argument index
    - Raises ValueError if the function does not return a single scalar value
    - Differentiating through non-differentiable operations (like ``abs`` at ``x=0``)
      may return a valid but potentially undefined result

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Basic example with scalar function
    >>> def f(x):
    >>>     return np.sin(x**2)
    >>>
    >>> df = arc.grad(f)
    >>> print(np.allclose(df(1.0), 2.0 * np.cos(1.0)))
    True
    >>>
    >>> # Multi-argument function with gradient w.r.t. specific argument
    >>> def loss(x, A, y):
    >>>     y_pred = A @ x
    >>>     return np.sum((y_pred - y)**2) / len(y)
    >>>
    >>> # Get gradient with respect to weights only
    >>> grad_A = arc.grad(loss, argnums=1)
    >>>
    >>> A = np.random.randn(2, 3)
    >>> x = np.ones(3)
    >>> y = np.array([1, 2])
    >>> print(grad_A(x, A, y))  # df/dA
    >>>
    >>> # Alternatively, differentiate with respect to multiple arguments
    >>> grads = arc.grad(loss, argnums=(0, 1))
    >>> # Returns a tuple (df/dx, df/dA)
    >>> print(grads(x, A, y))

    See Also
    --------
    jac : Compute the Jacobian matrix of a function
    hess : Compute the Hessian matrix of a scalar function
    """

    if not isinstance(func, FunctionCache):
        func = FunctionCache(
            func,
            static_argnums=static_argnums,
            static_argnames=static_argnames,
        )

    if isinstance(argnums, int):
        argnums = (argnums,)

    if not isinstance(argnums, tuple) or not all(isinstance(i, int) for i in argnums):
        raise ValueError("argnums must be an integer or a tuple of integers")

    if any(i in func.static_argnums for i in argnums):
        raise ValueError("Cannot differentiate with respect to a static argument")

    # Function to evaluate the gradient using the underlying CasADi function,
    # assuming that the arguments are already symbolic arrays. This can then
    # be used to create the gradient FunctionCache.
    def _grad(*args):
        # First make sure that the primal function has been compiled for these
        # argument types
        y = func(*args)
        if not isinstance(y, SymbolicArray):
            raise ValueError(
                "The primal function must return a single array. Multiple "
                "returns are not yet supported.  Return from "
                f"{func.name} is {y}"  # type: ignore[attr-defined]
            )
        if y.shape != ():
            raise ValueError(
                "The primal function must return a scalar value with shape (). "
                f"Return from {func.name} is {y} with shape "  # type: ignore[attr-defined]
                f"{y.shape}."
            )

        return tuple(y.grad(args[i]) for i in argnums)  # type: ignore[union-attr]

    if name is None:
        name = f"grad_{func.name}"

    _grad.__name__ = name

    return FunctionCache(
        _grad,
        arg_names=func.arg_names,
        static_argnums=func.static_argnums,
        kind=func._kind,
    )


def jac(
    func: Callable,
    argnums: int | Sequence[int] = 0,
    name: str | None = None,
    static_argnums: int | Sequence[int] | None = None,
    static_argnames: str | Sequence[str] | None = None,
) -> Callable:
    """Create a function that evaluates the Jacobian of ``func``.

    Transforms a vector-valued function into a new function that computes
    the Jacobian matrix (matrix of all first-order partial derivatives) with
    respect to one or more of its input arguments.

    Parameters
    ----------
    func : callable
        The function to differentiate. Can return a vector or matrix output.
        If not already a compiled function, it will be compiled with the
        specified static arguments.
    argnums : int or tuple of ints, optional
        Specifies which positional argument(s) to differentiate with respect to.
        Default is 0, meaning the first argument.
    name : str, optional
        Name for the created Jacobian function. If ``None``, a name is automatically
        generated based on the primal function's name.
    static_argnums : tuple of int, optional
        Specifies which positional arguments should be treated as static (not
        differentiated or traced symbolically). Only used if ``func`` is not already
        a compiled function.
    static_argnames : tuple of str, optional
        Specifies which keyword arguments should be treated as static. Only used
        if ``func`` is not already a compiled function.

    Returns
    -------
    callable
        A function that computes the Jacobian of ``func`` with respect to the
        specified arguments. If multiple arguments are specified in ``argnums``,
        the function returns a tuple of Jacobians, one for each specified argument.

    Notes
    -----
    When to use this function:

    - When working with derivatives of vector-valued functions
    - For constrained optimization problems where Jacobians are needed
    - For sensitivity analysis of vector outputs with respect to input parameters
    - For linearization of nonlinear models around operating points

    In cases where the full Jacobian is not needed, but only the product of the
    Jacobian with a vector, consider using :py:func:`jvp` (Jacobian-vector product) or
    :py:func:`vjp` (vector-transpose-Jacobian product) for more efficient computation.

    Currently this function only supports creating Jacobians for functions
    with a single return value. If the function has multiple return values,
    the function will raise a ``ValueError``.

    Internally, CasADi chooses between forward and reverse mode automatic
    differentiation using a heuristic based on the number of required derivative
    calculations in either case. For functions with many inputs and few outputs,
    reverse mode is typically more efficient. For functions with few inputs and
    many outputs, forward mode is typically preferred.

    Conceptual model:

    The Jacobian matrix represents the best linear approximation to a function
    near a given point. For a function :math:`f: R^n → R^m`, the Jacobian is an
    :math:`m \\times n` matrix where each element (i,j) represents the partial
    derivative of the i-th output with respect to the j-th input.  Hence, the function
    returned by ``jac`` takes the form :math:`J: R^n → R^{m \\times n}`.

    Edge cases:

    - Raises ``ValueError`` if ``argnums`` contains a static argument index
    - Raises ``ValueError`` if the function does not return a single array
    - Currently only supports functions with a single return value (future versions
      may support multiple returns)

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Example: Jacobian of a simple vector function
    >>> def f(x):
    >>>     return np.array([x[0]**2, x[0]*x[1], np.sin(x[1])], like=x)
    >>>
    >>> J = arc.jac(f)
    >>> x = np.array([2.0, 3.0])
    >>> print(J(x))
    [[ 4.         0.       ]
     [ 3.         2.       ]
     [ 0.        -0.9899925]]
    >>>
    >>> # Multi-argument function with Jacobian w.r.t. specific argument
    >>> def dynamics(t, x, u):
    >>>     # Simple pendulum dynamics with control input.
    >>>     g = 9.81
    >>>     L = 1.0
    >>>     return np.array([
    >>>         x[1], -(g * np.sin(x[0]) + u) / L
    >>>     ], like=x)
    >>>
    >>> # Get Jacobian with respect to state (for linearization)
    >>> Jx = arc.jac(dynamics, argnums=1)
    >>> t = 0.0
    >>> x0 = np.array([0.0, 0.0])  # Equilibrium
    >>> u = 0.0
    >>> print(Jx(t, x0, u))
    [[ 0.    1.  ]
     [-9.81  0.  ]]
    >>>
    >>> # Get Jacobian with respect to control input (for control design)
    >>> Ju = arc.jac(dynamics, argnums=2)
    >>> print(Ju(t, x0, u))
    [0. 1.]

    See Also
    --------
    grad : Compute the gradient of a scalar-valued function
    hess : Compute the Hessian matrix of a scalar function
    jvp : Compute Jacobian-vector products
    vjp : Compute vector-Jacobian products
    """
    # TODO: Support multiple returns using trees?

    if not isinstance(func, FunctionCache):
        func = FunctionCache(
            func,
            static_argnums=static_argnums,
            static_argnames=static_argnames,
        )

    if isinstance(argnums, int):
        argnums = (argnums,)

    if not isinstance(argnums, tuple) or not all(isinstance(i, int) for i in argnums):
        raise ValueError("argnums must be an integer or a tuple of integers")

    if any(i in func.static_argnums for i in argnums):
        raise ValueError("Cannot differentiate with respect to a static argument")

    # From the CasADi docs:
    # f: (x, y) -> (r, s) results in the function
    # df: (x, y, out_r, out_s) -> (jac_r_x, jac_r_y, jac_s_x, jac_s_y)

    # Function to evaluate the Jacobian using the underlying CasADi function,
    # assuming that the arguments are already symbolic arrays. This can then
    # be used to create the Jacobian FunctionCache.
    def _jac(*args):
        # First make sure that the primal function has been compiled for these
        # argument types
        y = func(*args)
        if not isinstance(y, SymbolicArray):
            raise ValueError(
                "The primal function must return a single array. Multiple "
                "returns are not yet supported.  Return from "
                f"{func.name} is {y}"  # type: ignore[attr-defined]
            )
        return tuple(y.jac(args[i]) for i in argnums)  # type: ignore[union-attr]

    if name is None:
        name = f"jac_{func.name}"

    _jac.__name__ = name

    return FunctionCache(
        _jac,
        arg_names=func.arg_names,
        static_argnums=func.static_argnums,
        kind=func._kind,
    )


def hess(
    func: Callable,
    argnums: int | Sequence[int] = 0,
    name: str | None = None,
    static_argnums: int | Sequence[int] | None = None,
    static_argnames: str | Sequence[str] | None = None,
) -> Callable:
    """Create a function that evaluates the Hessian of ``func``.

    Transforms a scalar-valued function into a new function that computes
    the Hessian matrix (matrix of all second-order partial derivatives) with
    respect to one or more of its input arguments.

    Parameters
    ----------
    func : callable
        The function to differentiate. Must be a scalar-valued function
        (returns a single value with shape ``()``). If not already a compiled
        function, it will be compiled with the specified static arguments.
    argnums : int or tuple of ints, optional
        Specifies which positional argument(s) to differentiate with respect to.
        Default is 0, meaning the first argument.
    name : str, optional
        Name for the created Hessian function. If None, a name is automatically
        generated based on the primal function's name.
    static_argnums : tuple of int, optional
        Specifies which positional arguments should be treated as static (not
        differentiated or traced symbolically). Only used if ``func`` is not already
        a compiled function.
    static_argnames : tuple of str, optional
        Specifies which keyword arguments should be treated as static. Only used
        if `func` is not already a compiled function.

    Returns
    -------
    callable
        A function that computes the Hessian of ``func`` with respect to the
        specified arguments. If multiple arguments are specified in ``argnums``,
        the function returns a tuple of Hessians, one for each specified argument.

    Notes
    -----
    When to use this function:

    - For optimization problems requiring second-derivative information
    - For analyzing the local curvature of a cost function
    - When working with quadratic approximations of nonlinear functions

    Conceptual model:

    The Hessian matrix represents the local curvature of a function. For a function
    :math:`f: R^n → R`, the Hessian is an :math:`n \\times n` symmetric matrix where
    each element (i,j) represents the second partial derivative
    :math:`\\partial^2 f / \\partial x_i \\partial x_j`.

    The Hessian is computed using automatic differentiation by applying the gradient
    operation twice. This ensures high numerical accuracy compared to finite
    differencing methods, especially for functions with complex computational paths.

    Edge cases:

    - Raises ``ValueError`` if ``argnums`` contains a static argument index
    - Currently only supports functions with a single return value

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Example: Hessian of a simple function
    >>> def f(x):
    >>>     return 100 * (x[1] - x[0]**2)**2 + (1 - x[0])**2  # Rosenbrock function
    >>>
    >>> H = arc.hess(f)
    >>> x = np.array([1.0, 1.0])  # At the minimum
    >>> print(H(x))
    [[ 802. -400.]
    [-400.  200.]]
    >>>
    >>> # Example: Hessian for optimization
    >>> def loss(x, A, y):
    >>>     y_pred = x @ A
    >>>     return np.sum((y_pred - y)**2) / len(y)
    >>>
    >>> # Get Hessian with respect to parameters
    >>> H_params = arc.hess(loss, argnums=1)
    >>>
    >>> # Create some example data
    >>> x = np.random.randn(100, 5)  # 100 samples, 5 features
    >>> y = np.random.randn(100)
    >>> A = np.zeros(5)  # Initial parameters
    >>>
    >>> # Compute Hessian at initial point (useful for Newton optimization)
    >>> H = H_params(x, A, y)
    >>> print(H.shape)  # Should be (5, 5)

    See Also
    --------
    grad : Compute the gradient of a scalar-valued function
    jac : Compute the Jacobian matrix of a function
    minimize : Optimization using automatic differentiation
    """
    # TODO: Support multiple returns using trees?

    if not isinstance(func, FunctionCache):
        func = FunctionCache(
            func,
            static_argnums=static_argnums,
            static_argnames=static_argnames,
        )

    if isinstance(argnums, int):
        argnums = (argnums,)

    if not isinstance(argnums, tuple) or not all(isinstance(i, int) for i in argnums):
        raise ValueError("argnums must be an integer or a tuple of integers")

    if any(i in func.static_argnums for i in argnums):
        raise ValueError("Cannot differentiate with respect to a static argument")

    def _hess(*args):
        # First make sure that the primal function has been compiled for these
        # argument types
        y = func(*args)
        if not isinstance(y, SymbolicArray):
            raise ValueError(
                "The primal function must return a single array. Multiple "
                "returns are not yet supported.  Return from "
                f"{func.name} is {y}"  # type: ignore[attr-defined]
            )
        if y.shape != ():
            raise ValueError(
                "The primal function must return a scalar value with shape (). "
                f"Return from {func.name} is {y} with shape "  # type: ignore[attr-defined]
                f"{y.shape}."
            )
        return tuple(y.hess(args[i]) for i in argnums)  # type: ignore[union-attr]

    if name is None:
        name = f"hess_{func.name}"

    _hess.__name__ = name

    return FunctionCache(
        _hess,
        arg_names=func.arg_names,
        static_argnums=func.static_argnums,
        kind=func._kind,
    )


def jvp(
    func: Callable,
    name: str | None = None,
    static_argnums: int | Sequence[int] | None = None,
    static_argnames: str | Sequence[str] | None = None,
) -> Callable:
    """Create a function that evaluates the Jacobian-vector product of ``func``.

    Transforms a function into a new function that efficiently computes the
    product of the Jacobian matrix of ``func`` with a given vector, using
    forward-mode automatic differentiation.

    Parameters
    ----------
    func : callable
        The function to differentiate. If not already a compiled function,
        it will be compiled with the specified static arguments.
    name : str, optional
        Name for the created JVP function. If ``None``, a name is automatically
        generated based on the primal function's name.
    static_argnums : tuple of int, optional
        Specifies which positional arguments should be treated as static (not
        differentiated or traced symbolically). Only used if ``func`` is not already
        a compiled function.
    static_argnames : tuple of str, optional
        Specifies which keyword arguments should be treated as static. Only used
        if ``func`` is not already a compiled function.

    Returns
    -------
    callable
        A function with signature ``jvp_fun(x, v)`` that computes
        :math:`J(x) \\cdot v`, where :math:`J(x)` is the Jacobian of ``func`` evaluated
        at :math:`x`, and :math:`v` is the vector to multiply with. The function
        returns a vector with the same shape as the output of ``func``.

    Notes
    -----
    When to use this function:

    - When you need directional derivatives along a specific vector
    - When computing sensitivities for functions with many outputs and few inputs
    - When the full Jacobian matrix would be too large to compute or store efficiently
    - In iterative algorithms that require repeated Jacobian-vector products

    Conceptual model:

    The Jacobian-vector product (JVP) computes the directional derivative of a
    function in the direction of a given vector, without explicitly forming the
    full Jacobian matrix. For a function :math:`f: R^n \\rightarrow R^m` and a vector
    :math:`v \\in R^n`, the JVP is equivalent to :math:`J(x) \\cdot v`, where
    :math:`J(x)` is the :math:`m \\times n` Jacobian matrix at point :math:`x`.

    Forward-mode automatic differentiation computes JVPs efficiently, with a
    computational cost similar to that of evaluating the original function,
    regardless of the output dimension. This makes JVP particularly effective for
    functions with few inputs but many outputs.

    The JVP also represents how a small change in the input (in the direction of
    :math:`v`) affects the output of the function, making it useful for sensitivity
    analysis.

    Edge cases:

    - Raises ``ValueError`` if the function does not return a single vector-valued\
        array
    - The vector ``v`` must have the same shape as the input ``x``

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Example: JVP of a simple function
    >>> def f(x):
    >>>     return np.array([x[0]**2, x[0]*x[1], np.exp(x[1])], like=x)
    >>>
    >>> # Create the JVP function
    >>> f_jvp = arc.jvp(f)
    >>>
    >>> # Evaluate at a point
    >>> x = np.array([2.0, 1.0])
    >>> v = np.array([1.0, 0.5])  # Direction vector
    >>>
    >>> # Compute JVP: J(x)·v
    >>> auto_jvp = f_jvp(x, v)
    >>> print(auto_jvp)
    [4.         2.         1.35914091]
    >>>
    >>> # Compare with direct computation of Jacobian
    >>> manual_jvp = arc.jac(f)(x) @ v
    >>> print(np.allclose(auto_jvp, manual_jvp))
    True
    >>>
    >>> # Example: Efficient sensitivity analysis for a high-dimensional output
    >>> def high_dim_func(params):
    >>>     # Function with few inputs but many outputs.
    >>>     return np.sin(np.sum(np.outer(params, np.arange(1000)), axis=0))
    >>>
    >>> # JVP is much more efficient than computing the full Jacobian
    >>> sensitivity = arc.jvp(high_dim_func)
    >>> params = np.array([0.1, 0.2])
    >>> direction = np.array([1.0, 0.0])  # Sensitivity in first parameter direction
    >>>
    >>> # Compute how output changes in the direction of the first parameter
    >>> output_sensitivity = sensitivity(params, direction)
    >>> print(output_sensitivity.shape)
    (1000,)

    See Also
    --------
    vjp : Compute vector-Jacobian products (reverse-mode AD)
    jac : Compute the full Jacobian matrix
    grad : Compute the gradient of a scalar-valued function
    """

    # Note that the interface here differs from JAX, which has the signature
    # `jvp(func, primals, tangents) -> primals, tangents`. In this case the JVP
    # is computed symbolically up front and can then be evaluated efficiently for
    # every primal/tangent pair.

    if not isinstance(func, FunctionCache):
        func = FunctionCache(
            func,
            static_argnums=static_argnums,
            static_argnames=static_argnames,
        )

    # Function to evaluate the JVP using the underlying CasADi function,
    # assuming that the arguments are already symbolic arrays. This can then
    # be used to create the JVP FunctionCache.
    def _jvp(x, v):
        # The return values can be a single SymbolicArray or a tuple of these.
        y = func(x)

        if not isinstance(y, SymbolicArray):
            raise ValueError(
                "The primal function must return a single array. Multiple "
                "returns are not yet supported.  Return from "
                f"{func.name} is {y}"  # type: ignore[attr-defined]
            )

        return y.jvp(x, v)

    if name is None:
        name = f"jvp_{func.name}"

    _jvp.__name__ = name

    func_name = func.name
    primal_name = func_name + "_primal"
    tangent_name = func_name + "_tangent"

    return FunctionCache(
        _jvp,
        arg_names=[primal_name, tangent_name],
        static_argnums=func.static_argnums,
        kind=func._kind,
    )


def vjp(
    func: Callable,
    name: str | None = None,
    static_argnums: int | Sequence[int] | None = None,
    static_argnames: str | Sequence[str] | None = None,
) -> Callable:
    """Create a function that evaluates the vector-Jacobian product of ``func``.

    Transforms a function into a new function that efficiently computes the
    product of a vector with the transposed Jacobian matrix of ``func``, using
    reverse-mode automatic differentiation.

    Parameters
    ----------
    func : callable
        The function to differentiate. If not already a compiled function,
        it will be compiled with the specified static arguments.
    name : str, optional
        Name for the created VJP function. If ``None``, a name is automatically
        generated based on the primal function's name.
    static_argnums : tuple of int, optional
        Specifies which positional arguments should be treated as static (not
        differentiated or traced symbolically). Only used if ``func`` is not already
        a compiled function.
    static_argnames : tuple of str, optional
        Specifies which keyword arguments should be treated as static. Only used
        if ``func`` is not already a compiled function.

    Returns
    -------
    callable
        A function with signature ``vjp_fun(x, w)`` that computes
        :math:`w^T \\cdot J(x)`, where :math:`J(x)` is the Jacobian of ``func``
        evaluated at :math:`x` and :math:`w` is the vector to multiply with.
        The function returns a vector with the same shape as the input ``x``.

    Notes
    -----
    When to use this function:

    - When you need gradients of scalar projections of the output
    - When computing sensitivities for functions with many inputs and few outputs
    - In adjoint-based (e.g PDE-constrained) optimization problems

    Conceptual model:

    The vector-Jacobian product (VJP) computes the gradient of a scalar projection
    of the output without explicitly forming the full Jacobian matrix. For a function
    :math:`f: R^n \\rightarrow R^m` and a vector :math:`w \\in R^m`, the VJP is
    equivalent to :math:`w^T \\cdot J(x)`, where :math:`J(x)` is the
    :math:`m \\times n` Jacobian matrix evaluated at point :math:`x`.

    Reverse-mode automatic differentiation computes VJPs efficiently, with a
    computational cost similar to that of evaluating the original function,
    regardless of the input dimension. This makes VJP particularly effective for
    functions with many inputs but few outputs (which is why it's widely used in
    machine learning for gradient-based optimization). If the function is scalar-valued
    then the VJP with vector ``w=1`` is equivalent to the gradient of the function.

    The VJP represents the sensitivity of a weighted sum of outputs to changes in
    each input dimension.

    Edge cases:

    - Raises ``ValueError`` if the function does not return a single vector-valued array
    - The vector ``w`` must have the same shape as the output of ``func``

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Example: VJP of a simple function
    >>> def f(x):
    >>>     return np.array([x[0]**2, x[0]*x[1], np.exp(x[1])], like=x)
    >>>
    >>> # Create the VJP function
    >>> f_vjp = arc.vjp(f)
    >>>
    >>> # Evaluate at a point
    >>> x = np.array([2.0, 1.0])
    >>> w = np.array([1.0, 0.5, 2.0])  # Weighting vector for outputs
    >>>
    >>> # Compute VJP: w^T·J(x)
    >>> auto_vjp = f_vjp(x, w)
    >>> print(auto_vjp)
    [4.5        6.43656366]
    >>>
    >>> # Compare with direct computation of Jacobian transpose
    >>> J = arc.jac(f)
    >>> full_jacobian = J(x)
    >>> manual_vjp = w @ full_jacobian
    >>> print(np.allclose(input_gradient, manual_vjp))
    True
    >>>
    >>> # Example: Comparing efficiency with JVP
    >>> # For functions with many inputs but few outputs, VJP is more efficient
    >>> def high_dim_input_func(x):
    >>>     # Sum of squares of many inputs, producing a scalar
    >>>     return np.sum(x**2)
    >>>
    >>> # Gradient computation is equivalent to a VJP with w=1
    >>> grad_result = arc.grad(high_dim_input_func)(x)
    >>> x = np.random.randn(1000)  # 1000-dimensional input, scalar output
    >>> w = np.array(1.0)  # Weight for the scalar output
    >>>
    >>> # VJP computes the gradient efficiently without forming full Jacobian
    >>> vjp_result = arc.vjp(high_dim_input_func)(x, w)
    >>> print(np.allclose(vjp_result, grad_result))
    True

    See Also
    --------
    jvp : Compute Jacobian-vector products (forward-mode AD)
    jac : Compute the full Jacobian matrix
    grad : Compute the gradient of a scalar-valued function (special case of VJP)
    """

    # Note that the interface here differs from JAX, which has the signature
    # `vjp(func, *primals) -> primals, vjp_fun`, where `vjp_fun` has the signature
    # `vjp_fun(cotangents) -> cotangents`. In this case the VJP is computed
    # symbolically up front and can then be evaluated efficiently for every
    # primal/cotangent pair.

    if not isinstance(func, FunctionCache):
        func = FunctionCache(
            func,
            static_argnums=static_argnums,
            static_argnames=static_argnames,
        )

    # For now, only support functions with a single argument and return value
    if len(func.arg_names) != 1:
        raise NotImplementedError("TODO: Support multiple arguments")

    def _vjp(x, w):
        # First make sure that the primal function has been compiled for these
        # argument types
        y = func(x)
        if not isinstance(y, SymbolicArray):
            raise ValueError(
                "The primal function must return a single array. Multiple "
                "returns are not yet supported.  Return from "
                f"{func.name} is {y}"  # type: ignore[attr-defined]
            )
        return y.vjp(x, w)

    if name is None:
        name = f"vjp_{func.name}"

    _vjp.__name__ = name

    # The new function will be evaluated with "primal" data that has the same shape
    # as the inputs, and "cotangent" data that has the same shape as the output. The
    # names need to match.
    no_return_names = func.return_names is None
    if no_return_names:
        # Since only a single return value is currently supported, we can just assume
        # that the function only outputs one value and give it an arbitrary name. If
        # the function actually returns multiple values, it will throw an error when
        # "compiled" later.
        func.return_names = ["y0"]
    cotangent_names = [f"d{name}" for name in func.return_names]

    # Reset to None so that the error about multiple returns is handled by
    # the VJP function and not by a mismatch in the number of return names.
    if no_return_names:
        func.return_names = None

    return FunctionCache(
        _vjp,
        arg_names=func.arg_names + cotangent_names,
        static_argnums=func.static_argnums,
        kind=func._kind,
    )



import numpy as np

from archimedes import compile, jac, odeint


def lqr_design(f, x0, u0, ts, Qf, Q, R):
    """Construct a finite-horizon LQR control law

    Args:
        f(t, x, u): dynamics function
        x0: nominal state (or function of time)
        u0: nominal control (or function of time)
        ts: time points for RDE solve
        Qf: final cost matrix
        Q: cost matrix
        R: control cost matrix
    """
    nx = Q.shape[0]
    nu = R.shape[0]
    R_inv = np.linalg.inv(R)

    if isinstance(x0, np.ndarray):

        def _x0(t):
            return x0
    else:
        _x0 = x0

    if isinstance(u0, np.ndarray):

        def _u0(t):
            return u0
    else:
        _u0 = u0

    # Linearize the dynamics about the nominal trajectory
    dfdx = jac(f, argnums=1)
    dfdu = jac(f, argnums=2)

    def linearize(t):
        x0, u0 = _x0(t), _u0(t)
        A = dfdx(t, x0, u0)
        B = dfdu(t, x0, u0)
        return A, B

    # Differential Riccatti equation (square root form)
    # Integrate Riccati equation from t = tf to t = 0
    # Transform time variable to tau = -t and
    # integrate from tau = -t0 to tau = -t0
    # In this case we just change the sign of S_t (usually has a negative sign
    # when integrating backwards in time)
    @compile(kind="MX")
    def riccati_rhs(t, x):
        t = -t  # Transform from tau=-t to t
        A, B = linearize(t)
        P = np.reshape(x, (nx, nx))
        S = P @ P.T
        P_t = A.T @ P - 0.5 * S @ (B @ R_inv @ B.T) @ P + 0.5 * Q * np.linalg.inv(P)
        return P_t.flatten()

    P0 = np.linalg.cholesky(Qf)
    t0, tf = ts[0], ts[-1]
    P = odeint(riccati_rhs, x0=P0.flatten(), t_span=(-tf, -t0), t_eval=ts)
    P = np.reshape(P.T, (len(ts), nx, nx))
    S = np.array([P[i] @ P[i].T for i in range(len(ts))])

    Ks = np.zeros((len(ts), nu, nx))
    for i in range(len(ts)):
        _A, B = linearize(ts[i])
        Ks[i] = R_inv @ B.T @ S[i]

    def K(t):
        # TODO: Use barycentric interpolation here?
        # Or something that doesn't require repeated interpolation
        _K = [np.interp(t, ts, Ks[:, i, j]) for i in range(nu) for j in range(nx)]
        return np.stack(_K).reshape((nu, nx))

    def u_lqr(t, x):
        return -K(t) @ (x - _x0(t))

    return u_lqr



# ruff: noqa: N802, N803, N806, E741

import numpy as np
import pytest

import archimedes as arc
from archimedes import discretize, struct
from archimedes.observers import ExtendedKalmanFilter
from archimedes.sysid import Timeseries, pem

np.random.seed(0)


@struct
class CartPole:
    """CartPole model for testing system identification."""

    m1: float = 1.0  # cart mass
    m2: float = 0.3  # pole mass
    L: float = 0.5  # pole length
    g: float = 9.81  # gravity

    def dynamics(self, t, q, u=None):
        """CartPole dynamics: [x, θ, ẋ, θ̇]"""
        x, θ, ẋ, θ̇ = q
        sθ, cθ = np.sin(θ), np.cos(θ)
        τ = 0.0 if u is None else np.atleast_1d(u)[0]
        den = self.m1 + self.m2 * sθ**2
        ẍ = (self.L * self.m2 * sθ * θ̇**2 + τ + self.m2 * self.g * cθ * sθ) / den
        θ̈ = -(
            self.L * self.m2 * cθ * sθ * θ̇**2
            + τ * cθ
            + (self.m1 + self.m2) * self.g * sθ
        ) / (self.L * den)
        return np.stack([ẋ, θ̇, ẍ, θ̈])


options = {
    "bfgs": None,
    "lm": None,
    "ipopt": None,
}


def second_order_ode(t, x, u, params):
    """Second-order system ODE: ẍ + 2ζωₙẋ + ωₙ²x = ωₙ²u"""
    omega_n = params["omega_n"]
    zeta = params["zeta"]

    x1, x2 = x[0], x[1]
    u_val = u[0]

    x1_t = x2
    x2_t = -(omega_n**2) * x1 - 2 * zeta * omega_n * x2 + omega_n**2 * u_val

    return np.hstack([x1_t, x2_t])


def position_obs(t, x, u, params):
    return x[0]


# Fixture for second-order ODE data
@pytest.fixture(scope="session")
def second_order_data():
    # True system parameters
    omega_n_true = 2.0  # rad/s
    zeta_true = 0.1  # damping ratio
    params_true = {"omega_n": omega_n_true, "zeta": zeta_true}

    # Time vector
    t0, tf = 0.0, 5.0
    dt = 0.05
    ts = np.arange(t0, tf, dt)

    # Problem dimensions
    nu = 1  # input dimension (u)
    ny = 1  # output dimension (y = x₁)

    # Input signal (step input)
    us = np.ones((nu, len(ts)))

    # Initial conditions
    x0_true = np.array([0.0, 0.0])  # start at rest

    # Generate reference data
    xs_true = arc.odeint(
        second_order_ode,
        t_span=(t0, tf),
        x0=x0_true,
        args=(us[:, 0], params_true),
        t_eval=ts,
        rtol=1e-8,
        atol=1e-10,
    )

    # Add small amount of measurement noise
    noise_std = 0.01
    ys = xs_true[:1, :] + np.random.normal(0, noise_std, (ny, len(ts)))

    data = Timeseries(ts=ts, us=us, ys=ys)

    return {
        "data": data,
        "params_true": params_true,
        "x0_true": x0_true,
        "xs_true": xs_true,
        "noise_std": noise_std,
    }


class TestPEMHarmonicOscillator:
    @pytest.mark.parametrize("method", options.keys())
    def test_second_order_system(self, method, second_order_data, plot=False):
        """Test parameter recovery on a second-order damped oscillator.

        System: ẍ + 2ζωₙẋ + ωₙ²x = ωₙ²u
        State space:
            ẋ₁ = x₂
            ẋ₂ = -ωₙ²x₁ - 2ζωₙx₂ + ωₙ²u
        Parameters:
            ωₙ (natural frequency)
            ζ (damping ratio)
        """

        # Problem dimensions
        nx = 2  # state dimension (x₁, x₂)
        ny = 1  # output dimension (y = x₁)

        # Extract data from fixture
        data = second_order_data["data"]
        params_true = second_order_data["params_true"]
        omega_n_true = params_true["omega_n"]
        zeta_true = params_true["zeta"]
        x0_true = second_order_data["x0_true"]
        xs_true = second_order_data["xs_true"]
        noise_std = second_order_data["noise_std"]
        dt = data.ts[1] - data.ts[0]  # time step from data

        # Initial parameter guess (should be different from true values)
        params_guess = {"omega_n": 2.5, "zeta": 0.5}

        R = noise_std**2 * np.eye(ny)  # Measurement noise covariance
        Q = noise_std**2 * np.eye(nx)  # Process noise covariance

        # Initial state covariance (not necessary, just tests proper handling)
        P0 = np.eye(nx)

        # Set up PEM problem
        dyn = discretize(second_order_ode, dt, method="rk4")
        ekf = ExtendedKalmanFilter(dyn, position_obs, Q, R)

        # Set up reasonable bounds (not necessary for convergence, just included
        # for testing purposes)
        bounds = (
            {"omega_n": 0.0, "zeta": 0.0},  # lower bounds
            {"omega_n": 10.0, "zeta": 1.0},  # upper bounds
        )

        result = pem(
            ekf,
            data,
            params_guess,
            x0=x0_true,  # Assume initial conditions are known
            P0=P0,
            bounds=bounds,
            method=method,
            options=options[method],
        )
        params_opt = result.p

        # Validate results
        print("\nSecond-Order System ID Results:")
        print(f"True parameters: ωₙ={omega_n_true:.3f}, ζ={zeta_true:.3f}")
        print(
            f"Estimated parameters: ωₙ={params_opt['omega_n']:.3f}, "
            f"ζ={params_opt['zeta']:.3f}"
        )
        print(f"Success: {result.success}")
        print(f"Iterations: {result.nit}")
        print(f"Final cost: {0.5 * np.dot(result.fun, result.fun):.2e}")

        # Validate forward simulation accuracy
        xs_pred = arc.odeint(
            second_order_ode,
            t_span=(data.ts[0], data.ts[-1]),
            x0=x0_true,
            args=(data.us[:, 0], params_opt),
            t_eval=data.ts,
            rtol=1e-8,
            atol=1e-10,
        )

        if plot:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(2, 1, figsize=(7, 4), sharex=True)
            ax[0].plot(data.ts, data.ys[0], label="Measured Output (y₁)")
            ax[0].plot(
                data.ts, xs_true[0], label="True Output (x₁)", c="k", linestyle="--"
            )
            ax[0].plot(data.ts, xs_pred[0], label="Final predictions", linestyle="-.")
            ax[0].legend()
            ax[0].grid()
            ax[0].set_ylabel("State prediction")
            # ax[1].plot(ts, kf_result_init["e"].T, label="Initial residuals")
            # ax[1].plot(ts, kf_result_opt["e"].T, label="Final residuals")
            ax[1].set_ylabel("Kalman residuals")
            ax[1].grid()
            ax[-1].set_xlabel("Time (s)")
            plt.show()

        # Test assertions
        assert result.success, f"Parameter estimation failed: {result.message}"

        # Check parameter recovery accuracy
        omega_n_error = abs(params_opt["omega_n"] - omega_n_true) / omega_n_true
        zeta_error = abs(params_opt["zeta"] - zeta_true) / zeta_true

        assert omega_n_error < 0.01, (
            f"Natural frequency error too large: {100 * omega_n_error:.6f} %"
        )
        assert zeta_error < 0.1, (
            f"Damping ratio error too large: {100 * zeta_error:.6f} %"
        )

        simulation_error = np.sqrt(np.mean((xs_true - xs_pred) ** 2))
        print(f"Forward simulation RMS error: {simulation_error:.2e}")

        assert simulation_error < 0.05, (
            f"Forward simulation error too large: {simulation_error:.6f}"
        )

        # Test convergence performance
        assert result.nit < 50, f"Too many iterations required: {result.nit}"
        final_cost = 0.5 * np.dot(result.fun, result.fun) / len(data)
        assert final_cost / len(data) < 1e-3, f"Final cost too high: {result.fun:.2e}"

    @pytest.mark.parametrize("method", options.keys())
    def test_multi_experiment(self, method, second_order_data):
        # Problem dimensions
        nx = 2  # state dimension (x₁, x₂)
        ny = 1  # output dimension (y = x₁)

        # Extract data from fixture
        data = second_order_data["data"]
        params_true = second_order_data["params_true"]
        omega_n_true = params_true["omega_n"]
        zeta_true = params_true["zeta"]
        x0_true = second_order_data["x0_true"]
        noise_std = second_order_data["noise_std"]
        dt = data.ts[1] - data.ts[0]  # time step from data

        # Initial parameter guess (should be different from true values)
        params_guess = {"omega_n": 2.5, "zeta": 0.5}

        R = noise_std**2 * np.eye(ny)  # Measurement noise covariance
        Q = noise_std**2 * np.eye(nx)  # Process noise covariance

        # Set up PEM problem
        dyn = discretize(second_order_ode, dt, method="rk4")
        ekf = ExtendedKalmanFilter(dyn, position_obs, Q, R)

        result = pem(
            ekf,
            [data, data],
            params_guess,
            x0=[x0_true, x0_true],
            method=method,
            options=options[method],
        )
        params_opt = result.p

        # Test assertions
        assert result.success, f"Parameter estimation failed: {result.message}"

        # Check parameter recovery accuracy
        omega_n_error = abs(params_opt["omega_n"] - omega_n_true) / omega_n_true
        zeta_error = abs(params_opt["zeta"] - zeta_true) / zeta_true

        assert omega_n_error < 0.01, (
            f"Natural frequency error too large: {100 * omega_n_error:.6f} %"
        )
        assert zeta_error < 0.1, (
            f"Damping ratio error too large: {100 * zeta_error:.6f} %"
        )

    @pytest.mark.parametrize("method", options.keys())
    def test_optimize_ics(self, method, second_order_data):
        # Problem dimensions
        nx = 2  # state dimension (x₁, x₂)
        ny = 1  # output dimension (y = x₁)

        # Extract data from fixture
        data = second_order_data["data"]
        params_true = second_order_data["params_true"]
        omega_n_true = params_true["omega_n"]
        zeta_true = params_true["zeta"]
        x0_true = second_order_data["x0_true"]
        noise_std = second_order_data["noise_std"]
        dt = data.ts[1] - data.ts[0]  # time step from data

        # Initial parameter guess (should be different from true values)
        params_guess = {"omega_n": 2.5, "zeta": 0.5}

        # Set up PEM problem
        R = noise_std**2 * np.eye(ny)  # Measurement noise covariance
        Q = (0.01 * noise_std) ** 2 * np.eye(nx)  # Process noise covariance

        dyn = discretize(second_order_ode, dt, method="rk4")
        ekf = ExtendedKalmanFilter(dyn, position_obs, Q, R)

        # Test optimizing initial conditions
        result = pem(
            ekf,
            data,
            params_guess,
            x0=np.array([1.0, 0.0]),
            estimate_x0=True,
            method=method,
            options=options[method],
        )
        x0_est, params_est = result.x0, result.p
        print(f"Estimated initial conditions: {x0_est}")
        print(f"Estimated parameters with x0: {params_est}")

        assert result.success, f"Parameter estimation with x0 failed: {result.message}"

        # Check parameter recovery accuracy (should be good for this clean problem)
        omega_n_error = abs(params_est["omega_n"] - omega_n_true)
        zeta_error = abs(params_est["zeta"] - zeta_true)

        assert omega_n_error < 0.02, (
            f"Natural frequency error too large: {omega_n_error:.6f}"
        )
        assert zeta_error < 0.01, f"Damping ratio error too large: {zeta_error:.6f}"

        assert np.allclose(x0_est[0], x0_true[0], atol=0.1), (
            f"Initial condition error too large: {np.abs(x0_est - x0_true)}"
        )

        # Error handling
        with pytest.raises(ValueError, match=r"Unsupported method.*"):
            pem(
                ekf,
                data,
                params_guess,
                x0=x0_true,
                method="unsupported_method",
            )


class TestPEMVanDerPol:
    def test_van_der_pol(self, plot=False):
        """Test parameter recovery on Van der Pol oscillator (nonlinear system).

        System: ẍ - μ(1-x²)ẋ + x = 0
        State space:
            ẋ₁ = x₂
            ẋ₂ = μ(1-x₁²)x₂ - x₁
        Parameters:
            μ (nonlinearity parameter)
        """
        # True system parameters
        mu_true = 0.5  # Moderate nonlinearity for good convergence
        params_true = {"mu": mu_true}

        # Time vector (shorter simulation for limit cycle development)
        t0, tf = 0.0, 15.0
        dt = 0.05
        ts = np.arange(t0, tf, dt)

        # Problem dimensions
        nx = 2  # state dimension (x₁, x₂)
        ny = 1  # output dimension (y = x₁)

        # No input signal (autonomous system)
        us = np.zeros((1, len(ts)))  # dummy input

        # Initial conditions (start away from equilibrium to excite dynamics)
        x0_true = np.array([2.0, 0.0])  # initial displacement

        # Generate true system response
        def van_der_pol_ode(t, x, u, params):
            """Van der Pol oscillator: ẍ - μ(1-x²)ẋ + x = 0"""
            mu = params["mu"]

            x1, x2 = x[0], x[1]
            # u is unused for autonomous system

            x1_t = x2
            x2_t = mu * (1 - x1**2) * x2 - x1

            return np.hstack([x1_t, x2_t])

        def obs(t, x, u, params):
            return x[0]  # observe position only

        # Generate reference data
        xs_true = arc.odeint(
            van_der_pol_ode,
            t_span=(t0, tf),
            x0=x0_true,
            args=(us[:, 0], params_true),
            t_eval=ts,
            rtol=1e-8,
            atol=1e-10,
        )

        # Add measurement noise
        noise_std = 0.01
        ys = xs_true[:1, :] + np.random.normal(0, noise_std, (ny, len(ts)))

        # Initial parameter guess (should be different from true value)
        params_guess = {"mu": 1.0}  # Start with higher nonlinearity

        R = noise_std**2 * np.eye(ny)  # Measurement noise covariance
        Q = noise_std**2 * np.eye(nx)  # Process noise covariance (scale with R)

        P0 = 1e-4 * np.eye(nx)  # Initial state covariance (small uncertainty)

        # Set up PEM problem
        dyn = discretize(van_der_pol_ode, dt, method="rk4")
        ekf = ExtendedKalmanFilter(dyn, obs, Q, R)
        data = Timeseries(ts=ts, us=us, ys=ys)
        result = pem(
            ekf,
            data,
            params_guess,
            x0=x0_true,  # Assume initial conditions are known
            method="bfgs",
        )
        params_opt = result.p

        # Validate results
        print("\nVan der Pol Oscillator ID Results:")
        print(f"True parameter: μ={mu_true:.3f}")
        print(f"Estimated parameter: μ={params_opt['mu']:.3f}")
        print(f"Success: {result.success}")
        print(f"Iterations: {result.nit}")
        print(f"Final cost: {result.fun:.2e}")

        # Validate forward simulation accuracy
        xs_pred = arc.odeint(
            van_der_pol_ode,
            t_span=(t0, tf),
            x0=x0_true,
            args=(us[:, 0], params_opt),
            t_eval=ts,
            rtol=1e-8,
            atol=1e-10,
        )

        if plot:
            import matplotlib.pyplot as plt

            pem_obj = arc.sysid.PEMObjective(ekf, data, P0, x0=x0_true)

            kf_result_init = pem_obj.forward(x0_true, params_guess)
            kf_result_opt = pem_obj.forward(x0_true, params_opt)

            fig, ax = plt.subplots(3, 1, figsize=(8, 6), sharex=True)

            # State trajectory
            ax[0].plot(ts, ys[0], label="Measured Output (x₁)", alpha=0.7)
            ax[0].plot(ts, xs_true[0], label="True Output (x₁)", c="k", linestyle="--")
            ax[0].plot(
                ts, kf_result_init["x_hat"][0], label="KF Estimate (x₁)", linestyle=":"
            )
            ax[0].plot(ts, xs_pred[0], label="Final predictions", linestyle="-.")
            ax[0].legend()
            ax[0].grid()
            ax[0].set_ylabel("Position (x₁)")
            ax[0].set_title("Van der Pol Oscillator - State Trajectories")

            # Kalman residuals
            ax[1].plot(ts, kf_result_init["e"].T, label="Initial residuals", alpha=0.7)
            ax[1].plot(ts, kf_result_opt["e"].T, label="Final residuals", alpha=0.7)
            ax[1].set_ylabel("Kalman residuals")
            ax[1].legend()
            ax[1].grid()

            # Phase portrait
            ax[2].plot(
                xs_true[0],
                xs_true[1],
                label="True phase portrait",
                c="k",
                linestyle="--",
            )
            ax[2].plot(
                xs_pred[0], xs_pred[1], label="Predicted phase portrait", linestyle="-."
            )
            ax[2].set_xlabel("Position (x₁)")
            ax[2].set_ylabel("Velocity (x₂)")
            ax[2].legend()
            ax[2].grid()
            ax[2].set_title("Phase Portrait")

            plt.tight_layout()
            plt.show()

        # Test assertions
        assert result.success, f"Parameter estimation failed: {result.message}"

        # Check parameter recovery accuracy
        # Note: Nonlinear systems are typically harder to identify than linear ones
        mu_error = abs(params_opt["mu"] - mu_true)

        assert mu_error < 0.05, (
            f"Nonlinearity parameter error too large: {mu_error:.6f}"
        )

        # Check simulation accuracy (allow larger error for nonlinear system)
        simulation_error = np.sqrt(np.mean((xs_true - xs_pred) ** 2))
        print(f"Forward simulation RMS error: {simulation_error:.2e}")

        assert simulation_error < 0.1, (
            f"Forward simulation error too large: {simulation_error:.6f}"
        )

        # Test convergence performance (allow more iterations for nonlinear system)
        assert result.nit < 100, f"Too many iterations required: {result.nit}"
        assert result.fun < 1e-2, f"Final cost too high: {result.fun:.2e}"


class TestPEMCartPole:
    def test_cartpole(self, plot=False):
        """Test parameter recovery on CartPole system (nonlinear underactuated system).

        System: CartPole with cart mass m1, pole mass m2, and pole length L
        State: [x, θ, ẋ, θ̇] (cart position, pole angle, velocities)
        Input: u (horizontal force on cart)
        Parameters to identify:
            m1 (cart mass)
            m2 (pole mass)
            L (pole length)
        Fixed parameter:
            g (gravity - assumed known)
        """
        # True system parameters (to be identified)
        m1_true = 1.0  # cart mass (kg)
        m2_true = 0.3  # pole mass (kg)
        L_true = 0.5  # pole length (m)
        g_true = 9.81  # gravity (m/s²) - assumed known

        params_true = {"m2": m2_true, "L": L_true}
        system = CartPole(
            **params_true, m1=m1_true, g=g_true
        )  # Create true system instance

        # Time vector (shorter simulation for challenging nonlinear system)
        t0, tf = 0.0, 3.0
        dt = 0.01  # Fine timestep for accurate simulation
        ts = np.arange(t0, tf, dt)

        # Problem dimensions
        nx = 4  # state dimension [x, θ, ẋ, θ̇]
        nu = 1  # input dimension (force on cart)
        ny = 4  # output dimension

        # Input signal
        us = np.zeros((nu, len(ts)))

        # Initial conditions (start near inverted equilibrium)
        x0_true = np.array([0.0, 0.1, 0.0, 0.0])  # Small initial pole angle

        # Generate true system response
        def cartpole_ode(t, x, u, params):
            return system.replace(**params).dynamics(t, x, u)

        def obs(t, x, u, params):
            return x

        def ode_rhs(t, x, params):
            """ODE with interpolated input."""
            u = np.interp(t, ts, us[0]).reshape((nu,))
            return cartpole_ode(t, x, u, params)

        # Generate reference data
        xs_true = arc.odeint(
            ode_rhs,
            t_span=(t0, tf),
            x0=x0_true,
            args=(params_true,),
            t_eval=ts,
            rtol=1e-8,
            atol=1e-10,
        )

        # Add measurement noise
        noise_std = 0.01
        ys = xs_true + np.random.normal(0, noise_std, (ny, len(ts)))

        # Initial parameter guess
        params_guess = {
            "m2": 0.5,
            "L": 1.0,
        }

        R = noise_std**2 * np.eye(ny)  # Measurement noise covariance
        Q = noise_std**2 * np.eye(nx)  # Process noise (smaller than measurement)

        # Set up PEM problem
        dyn = discretize(cartpole_ode, dt, method="rk4")
        ekf = ExtendedKalmanFilter(dyn, obs, Q, R)
        data = Timeseries(ts=ts, us=us, ys=ys)
        result = pem(
            ekf,
            data,
            params_guess,
            x0=x0_true,  # Assume initial conditions are known
            method="bfgs",
        )
        params_opt = result.p

        # Validate results
        print("\nCartPole System ID Results:")
        print(f"True parameters: m1={m1_true:.3f}, m2={m2_true:.3f}, L={L_true:.3f}")
        print(
            f"Estimated parameters: m2={params_opt['m2']:.3f}, L={params_opt['L']:.3f}"
        )
        print(f"Success: {result.success}")
        print(f"Iterations: {result.nit}")
        print(f"Final cost: {result.fun:.2e}")

        # Validate forward simulation accuracy
        xs_pred = arc.odeint(
            ode_rhs,
            t_span=(t0, tf),
            x0=x0_true,
            args=(params_opt,),
            t_eval=ts,
            rtol=1e-8,
            atol=1e-10,
        )

        if plot:
            import matplotlib.pyplot as plt

            P0 = 1e-4 * np.eye(nx)  # Initial state covariance (small uncertainty)
            pem_obj = arc.sysid.PEMObjective(ekf, data, P0, x0=x0_true)

            kf_result_init = pem_obj.forward(x0_true, params_guess)
            kf_result_opt = pem_obj.forward(x0_true, params_opt)

            fig, ax = plt.subplots(4, 1, figsize=(10, 8), sharex=True)

            # Cart position
            ax[0].plot(ts, ys[0], label="Measured Cart Position", alpha=0.7)
            ax[0].plot(
                ts, xs_true[0], label="True Cart Position", c="k", linestyle="--"
            )
            ax[0].plot(ts, xs_pred[0], label="Predicted Cart Position", linestyle="-.")
            ax[0].set_ylabel("Cart Position (m)")
            ax[0].legend()
            ax[0].grid()
            ax[0].set_title("CartPole System Identification Results")

            # Pole angle
            ax[1].plot(ts, ys[1], label="Measured Pole Angle", alpha=0.7)
            ax[1].plot(ts, xs_true[1], label="True Pole Angle", c="k", linestyle="--")
            ax[1].plot(ts, xs_pred[1], label="Predicted Pole Angle", linestyle="-.")
            ax[1].set_ylabel("Pole Angle (rad)")
            ax[1].legend()
            ax[1].grid()

            # Control input
            ax[2].plot(ts, us[0], label="Control Force", c="orange")
            ax[2].set_ylabel("Force (N)")
            ax[2].legend()
            ax[2].grid()

            # Kalman residuals
            ax[3].plot(
                ts, kf_result_init["e"][0], label="Initial residuals (x)", alpha=0.7
            )
            ax[3].plot(
                ts, kf_result_opt["e"][0], label="Final residuals (x)", alpha=0.7
            )
            ax[3].plot(
                ts,
                kf_result_init["e"][1],
                label="Initial residuals (θ)",
                alpha=0.7,
                linestyle=":",
            )
            ax[3].plot(
                ts,
                kf_result_opt["e"][1],
                label="Final residuals (θ)",
                alpha=0.7,
                linestyle=":",
            )
            ax[3].set_ylabel("Kalman Residuals")
            ax[3].set_xlabel("Time (s)")
            ax[3].legend()
            ax[3].grid()

            plt.tight_layout()
            plt.show()

        # Test assertions
        assert result.success, f"Parameter estimation failed: {result.message}"

        # Check parameter recovery accuracy
        # CartPole is challenging due to underactuation and nonlinearity
        m2_error = abs(params_opt["m2"] - m2_true) / m2_true
        L_error = abs(params_opt["L"] - L_true) / L_true

        print(f"Relative parameter errors: m2={m2_error:.3f}, L={L_error:.3f}")

        # Allow reasonable tolerances for this challenging system
        assert m2_error < 0.05, f"Pole mass error too large: {m2_error:.6f}"
        assert L_error < 0.05, f"Pole length error too large: {L_error:.6f}"

        # Check simulation accuracy (focus on observed states)
        obs_error = np.sqrt(np.mean((xs_true[:2] - xs_pred[:2]) ** 2))
        print(f"Observed states RMS error: {obs_error:.3e}")

        assert obs_error < 0.05, f"Simulation error too large: {obs_error:.6f}"

        # Test convergence performance
        assert result.nit < 50, f"Too many iterations required: {result.nit}"
        assert result.fun < 1e-1, f"Final cost too high: {result.fun:.2e}"


if __name__ == "__main__":
    # Run individual tests for debugging

    print("=" * 60)
    print("Running System Identification Tests")
    print("=" * 60)

    for i in range(10):
        print(f"Iteration {i + 1}")
        np.random.seed(i)

        print("\n" + "=" * 60)
        print("Running Second Order System Identification Test")
        print("=" * 60)
        TestPEMHarmonicOscillator().test_second_order_system(plot=False)

    # test_suite.test_second_order_system(plot=True)

    # print("\n" + "=" * 60)
    # print("Running Van der Pol Oscillator Identification Test")
    # print("=" * 60)

    # test_suite.test_van_der_pol(plot=True)

    # print("\n" + "=" * 60)
    # print("Running Duffing Oscillator Identification Test")
    # print("=" * 60)

    # test_suite.test_duffing(plot=True)

    # print("\n" + "=" * 60)
    # print("Running CartPole System Identification Test")
    # print("=" * 60)

    # test_suite.test_cartpole(plot=True)

    print("\n" + "=" * 60)
    print("All tests passed!")
    print("=" * 60)



"""System identification and parameter estimation functionality"""

from ._pem import PEMObjective, pem
from ._timeseries import Timeseries

__all__ = [
    "Timeseries",
    "pem",
    "PEMObjective",
]



import matplotlib as mpl
from cycler import cycler

__all__ = ["color_schemes", "set_theme"]

dark_charcoal = "#2A2A2A"
copper_orange = "#D35400"
rich_brown = "#5D4037"
ember_red = "#C0392B"
pale_gold = "#F1C40F"
light_gray = "#F5F5F5"
slate_blue = "#34495E"
steel_blue = "#5C9DC0"

color_schemes = {
    "light": {
        "background": light_gray,
        "text": dark_charcoal,
        "grid": "#DDDDDD",
        "primary": copper_orange,
        "secondary": slate_blue,
        "tertiary": pale_gold,
        "highlight": ember_red,
    },
    "dark": {
        "background": dark_charcoal,
        "text": light_gray,
        "grid": "#444444",
        "primary": copper_orange,
        "secondary": steel_blue,
        "tertiary": pale_gold,
        "highlight": ember_red,
    },
}


def set_theme(theme: str):
    """
    Configure matplotlib plotting parameters with a predefined color theme.

    This function sets up matplotlib's global parameters (rcParams) to apply
    a consistent color scheme across all subsequently created plots. It configures
    colors for plot elements including backgrounds, text, grid lines, and the color
    cycle used for plotting multiple data series.

    All documentation and example notebooks should check for an environment variable
    `ARCHIMEDES_THEME` to set the theme; this is automatically set to build the
    light/dark themed documentation.

    Parameters
    ----------
    theme : str
        The name of the color theme to apply. Must be one of ('light', 'dark').

    Returns
    -------
    None
        This function modifies matplotlib's global state but does not return a value.

    Notes
    -----
    This function modifies matplotlib's global rcParams, which affects all subsequent
    plots created in the same session. Call this function before creating any plots
    that should use the specified theme.

    The function sets the following matplotlib parameters:
    - figure.figsize: Sets a default figure size of (7, 3)
    - Background colors (axes.facecolor, figure.facecolor)
    - Text colors (text.color, axes.labelcolor, xtick.color, ytick.color)
    - Grid colors (grid.color)
    - Color cycle for multiple series (axes.prop_cycle)

    Examples
    --------
    >>> import archimedes as arc
    >>> import matplotlib.pyplot as plt
    >>> arc.set_theme('dark')
    >>> plt.plot([1, 2, 3, 4])
    >>> plt.title('A plot with dark theme')
    >>> plt.show()

    >>> # Change to another theme
    >>> arc.set_theme('light')
    >>> plt.figure()
    >>> plt.plot([4, 3, 2, 1])
    >>> plt.title('A plot with light theme')
    >>> plt.show()

    See Also
    --------
    matplotlib.pyplot.rcParams : The parameter dictionary that this function modifies
    """

    colors = color_schemes[theme]

    # Set up the figure with theme colors
    mpl.rcParams["figure.figsize"] = (7, 3)
    mpl.rcParams["axes.facecolor"] = colors["background"]
    mpl.rcParams["figure.facecolor"] = colors["background"]
    mpl.rcParams["text.color"] = colors["text"]
    mpl.rcParams["axes.labelcolor"] = colors["text"]
    mpl.rcParams["xtick.color"] = colors["text"]
    mpl.rcParams["ytick.color"] = colors["text"]
    mpl.rcParams["grid.color"] = colors["grid"]

    mpl.rcParams["axes.prop_cycle"] = cycler(
        color=[
            colors["primary"],
            colors["secondary"],
            colors["tertiary"],
        ]
    )



"""Animate a steady level turn maneuver of the F-16.

Requires a low-poly F-16 STL file, available here:
https://www.printables.com/model/840061-low-poly-f-16-falcon-aka-viper-jet-fighter/files

For additional dependencies: `uv pip install meshio`
"""

# ruff: noqa: N803, N806, N816
from __future__ import annotations

from pprint import pprint

import matplotlib.pyplot as plt
import meshio
import numpy as np
from f16 import SubsonicF16
from matplotlib.animation import FuncAnimation, PillowWriter
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

import archimedes as arc


def traj_data(xs_flat, unravel, stride=5):
    xs = arc.vmap(unravel)(xs_flat.T)

    positions = xs.pos.copy()
    positions[:, 2] *= -1  # Use altitude as Z for visualization
    positions *= 1 / 1000  # Scale to kft

    orientations = []
    for i in range(len(ts)):
        x = unravel(xs_flat[:, i])
        # The rotation matrix is for a "passive" rotation (changing coordinate
        # systems).  For visualization we want the "active" rotation, which
        # will rotate the points in a single coordinate system. Hence we take the
        # transpose.
        orientations.append(x.att.as_matrix().T)

    def att_as_rpy(x_flat):
        x = unravel(x_flat)
        rpy = np.mod(x.att.as_euler("xyz"), 2 * np.pi)
        return np.rad2deg(rpy)

    rpy = arc.vmap(att_as_rpy, in_axes=1)(xs_flat)

    return {
        "positions": positions[::stride],
        "orientations": orientations[::stride],
        "rpy": rpy[::stride],
        "time": ts[::stride],
    }


def plot_f16_trajectory(traj_data, mesh_file):
    positions = traj_data["positions"]
    orientations = traj_data["orientations"]
    rpy = traj_data["rpy"]

    fig = plt.figure(figsize=(10, 8), layout="constrained")
    ax = fig.add_subplot(111, projection="3d")

    # Read the surface mesh
    mesh = meshio.read(mesh_file)
    vertices = mesh.points
    faces = mesh.cells[0].data

    vertices = (vertices - vertices.mean(axis=0)) / 50  # center and scale

    vertices = vertices[:, [2, 0, 1]]
    vertices[:, 2] *= -1  # Flip Z axis
    vertices[:, 1] *= -1  # Flip Y axis

    # Shift to origin at approximate CG
    vertices[:, 0] += 0.2
    vertices[:, 2] -= 0.1

    # Calculate bounds from trajectory
    x_min, x_max = positions[:, 0].min(), positions[:, 0].max()
    y_min, y_max = positions[:, 1].min(), positions[:, 1].max()
    z_min, z_max = positions[:, 2].min(), positions[:, 2].max()

    # Add some padding
    padding = 0.1
    x_pad = (x_max - x_min) * padding
    y_pad = (y_max - y_min) * padding
    z_pad = 2

    # # Set up the plot
    ax.set_xlim(x_min - x_pad, x_max + x_pad)
    ax.set_ylim(y_min - y_pad, y_max + y_pad)
    ax.set_zlim(z_min - z_pad, z_max + z_pad)
    ax.set_xlabel("N [kft]")
    ax.set_ylabel("E [kft]")
    ax.set_zlabel("alt [kft]")
    ax.set_aspect("equal")

    # Trail line
    (trail_line,) = ax.plot([], [], [], linewidth=2)

    # Text
    time_text = ax.text2D(0.05, 0.75, "", transform=ax.transAxes)

    limits = np.array([getattr(ax, f"get_{axis}lim")() for axis in "xyz"])
    ax.set_box_aspect(np.ptp(limits, axis=1))

    def update(frame):
        for coll in ax.collections:
            coll.remove()

        # Transform vertices
        transformed = vertices @ orientations[frame].T
        transformed[:, 2] *= -1  # Flip z back to altitude coordinates
        transformed += positions[frame]

        # Add aircraft mesh
        mesh = Poly3DCollection(
            transformed[faces],
            alpha=0.8,
            facecolor="dimgray",
            edgecolor="k",
            linewidth=0.2,
        )
        ax.add_collection3d(mesh)

        # Update trail
        trail_line.set_data(positions[: frame + 1, 0], positions[: frame + 1, 1])
        trail_line.set_3d_properties(positions[: frame + 1, 2])
        # trail_line.set_color()

        # Update text
        time_text.set_text(
            f"Roll: {rpy[frame, 0]:.1f}°\nPitch: {rpy[frame, 1]:.1f}°\n"
            f"Yaw: {rpy[frame, 2]:.1f}°"
        )

        return mesh, trail_line, time_text

    anim = FuncAnimation(
        fig,
        update,
        frames=len(positions),
        interval=50,
        blit=False,
    )
    plt.close()
    return anim


if __name__ == "__main__":
    # Steady turn rate
    model = SubsonicF16(xcg=0.3)
    result = model.trim(vt=500, turn_rate=0.1, alt=10000.0, gamma=0.0)
    pprint(result.variables)

    x0 = result.state
    u0 = result.inputs

    x0_flat, unravel = arc.tree.ravel(x0)

    def ode_rhs(t, x_flat):
        x = unravel(x_flat)
        x_t = model.dynamics(t, x, u0)
        x_t_flat, _ = arc.tree.ravel(x_t)
        return x_t_flat

    t0, tf = 0.0, 60.0
    ts = np.arange(t0, tf, 0.1)
    xs_flat = arc.odeint(ode_rhs, (t0, tf), x0_flat, t_eval=ts)
    xs = arc.vmap(unravel)(xs_flat.T)

    data = traj_data(xs_flat, unravel, stride=5)
    for mode in ("light", "dark"):
        arc.theme.set_theme(mode)
        anim = plot_f16_trajectory(data, "_static/f16.stl")
        anim.save(f"_static/f16_turn_{mode}.gif", writer=PillowWriter(fps=20))



# ruff: noqa: N806, N803

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, TypeVar

import numpy as np
from scipy.optimize import OptimizeResult
from scipy.optimize import minimize as scipy_minimize

import archimedes as arc
from archimedes import compile, scan, tree
from archimedes.optimize import lm_solve

if TYPE_CHECKING:
    from archimedes.observers import KalmanFilterBase
    from archimedes.typing import Tree

    from ._timeseries import Timeseries

    T = TypeVar("T", bound=Tree)


__all__ = ["pem"]


@tree.struct
class PEMObjective:
    """Prediction Error Minimization objective function for system identification.

    Low-level interface for PEM optimization that provides both residual and
    cost function evaluations. This class encapsulates the Kalman filter
    forward pass and automatic differentiation for gradient computation.

    This is typically used internally by :func:`pem`. For most applications,
    use the higher-level :func:`pem` function directly. This can be useful mainly
    for constructing custom optimization workflows.

    Parameters
    ----------
    predictor : KalmanFilterBase
        Kalman filter implementing the system model with
        ``step(t, x, y, P, args)`` method.
    data : Timeseries
        Input-output data with ``ts``, ``us``, and ``ys`` arrays.
    P0 : array_like
        Initial state covariance matrix of shape ``(nx, nx)``.
    x0 : array_like, optional
        Initial state estimate of shape ``(nx,)``. If None, the optimization
        variables should include both initial state and parameters.

    Notes
    -----
    The objective implements the prediction error formulation:

    .. code-block:: text

        J = (1/N) Σ[k=1 to N] e[k]ᵀ e[k]

    where ``e[k]`` are the Kalman filter innovations (prediction errors).

    See Also
    --------
    pem : High-level parameter estimation interface
    """

    predictor: KalmanFilterBase
    data: Timeseries
    P0: np.ndarray | None
    x0: Optional[np.ndarray] = tree.field(static=True, default=None)  # type: ignore

    def forward(self, x0: np.ndarray, params: Tree) -> dict:
        """Run Kalman filter forward pass and compute prediction errors.

        Parameters
        ----------
        x0 : array_like
            Initial state estimate of shape ``(nx,)``.
        params : Tree
            Model parameters with arbitrary nested structure.

        Returns
        -------
        dict
            Results dictionary with keys:

            - ``"x_hat"`` : State estimates of shape ``(nx, N)``
            - ``"e"`` : Prediction errors of shape ``(ny, N)``
            - ``"V"`` : Average cost function value (scalar)
        """
        ts = self.data.ts
        us = self.data.us
        ys = self.data.ys

        nx = self.predictor.nx
        nu = us.shape[0]

        if self.P0 is None:
            P0 = np.eye(nx)
        else:
            P0 = self.P0

        V = 0.0  # Cost function
        init_carry, unravel_carry = tree.ravel((x0, P0, params, V))

        @compile(kind="MX")
        def scan_fn(carry_flat, input):
            t, u, y = input[0], input[1 : nu + 1], input[nu + 1 :]
            x, P, params, V = unravel_carry(carry_flat)
            x, P, e = self.predictor.step(t, x, y, P, args=(u, params))
            output = np.concatenate([x, e], axis=0)

            # Accumulate cost function, Jacobian, and Hessian
            V += 0.5 * np.sum(e**2)

            carry, _ = tree.ravel((x, P, params, V))
            return carry, output

        inputs = np.vstack((ts, us, ys)).T
        carry, scan_output = scan(scan_fn, init_carry, xs=inputs)
        _, _, _, V = unravel_carry(carry)
        scan_output = scan_output.T
        x_hat, e = scan_output[:nx], scan_output[nx:]

        # Average the function results
        V /= ts.size

        return {
            "x_hat": x_hat,
            "e": e,
            "V": V,
        }

    def residuals(
        self, decision_variables: Tree | tuple[np.ndarray, Tree]
    ) -> np.ndarray:
        """Evaluate prediction error residuals for least-squares optimization.

        Parameters
        ----------
        decision_variables : Tree or tuple[np.ndarray, Tree]
            Optimization parameters. If ``x0`` is provided during construction,
            this contains only model parameters. Otherwise, it contains both
            initial state and parameters as a flattened array.

        Returns
        -------
        residuals : ndarray
            Flattened prediction errors of shape ``(ny * N,)`` suitable for
            least-squares optimization methods like Levenberg-Marquardt.
        """
        if self.x0 is not None:
            x0 = self.x0
            _, params = decision_variables  # Dummy x0 for interface compatibility
        else:
            x0, params = decision_variables

        results = self.forward(x0, params)
        e: np.ndarray = results["e"].flatten()
        return e

    def __call__(self, decision_variables: Tree | tuple[np.ndarray, Tree]) -> float:
        """Evaluate cost function for general optimization methods.

        Parameters
        ----------
        decision_variables : array_like
            Optimization parameters. If ``x0`` is provided during construction,
            this contains only model parameters. Otherwise, it contains both
            initial state and parameters as a flattened array.

        Returns
        -------
        cost : float
            Average prediction error cost function value.

        Notes
        -----
        This method provides the scalar cost function interface needed for
        general optimization methods (BFGS, IPOPT, etc.).
        """
        if self.x0 is not None:
            x0 = self.x0
            _, params = decision_variables  # Dummy x0 for interface compatibility
        else:
            x0, params = decision_variables

        results = self.forward(x0, params)
        V: float = results["V"]
        return V


@tree.struct
class CompoundPEMObjective:
    """Compound PEM objective function for multiple sub-objectives.

    Shares the same interface as `PEMObjective` but allows
    combining multiple PEM objectives into a single cost function.
    """

    sub_objectives: list[PEMObjective]

    def _make_sub_dvs(
        self, decision_variables: tuple[np.ndarray, Tree]
    ) -> list[tuple[np.ndarray | None, Tree]]:
        """Prepare sub-decision variables for each sub-objective."""
        sub_x0, params = decision_variables
        sub_dvs: list[tuple[np.ndarray | None, Tree]] = []

        for x0, pem_obj in zip(sub_x0, self.sub_objectives):
            if pem_obj.x0 is not None:
                # If x0 is provided, use it directly
                sub_dvs.append((None, params))

            else:
                # Otherwise, include x0 in the decision variables
                sub_dvs.append((x0, params))

        return sub_dvs

    def __call__(self, decision_variables: tuple[np.ndarray, Tree]) -> float:
        """Evaluate cost function for compound PEM objective."""
        sub_dvs = self._make_sub_dvs(decision_variables)
        total_cost = 0.0
        for dvs, pem_obj in zip(sub_dvs, self.sub_objectives):
            cost = pem_obj(dvs)
            total_cost += cost

        return total_cost

    def residuals(
        self, decision_variables: Tree | tuple[np.ndarray, Tree]
    ) -> np.ndarray:
        """Evaluate residuals for compound PEM objective."""
        sub_dvs = self._make_sub_dvs(decision_variables)
        total_residuals = []
        for dvs, pem_obj in zip(sub_dvs, self.sub_objectives):
            residuals = pem_obj.residuals(dvs)
            total_residuals.append(residuals)

        if len(total_residuals) == 1:
            return total_residuals[0]

        return np.hstack(total_residuals)


def _pem_solve_lm(
    pem_obj: CompoundPEMObjective,
    p_guess: T,
    bounds: tuple[T, T] | None = None,
    options: dict | None = None,
) -> OptimizeResult:
    """Solve the PEM problem using Levenberg-Marquardt optimization."""
    if options is None:
        options = {}

    # Set reasonable defaults for system identification problems
    default_options = {
        "ftol": 1e-4,
        "xtol": 1e-6,
        "gtol": 1e-6,
        "max_nfev": 200,
    }
    options = {**default_options, **options}
    result: OptimizeResult = lm_solve(
        pem_obj.residuals, p_guess, bounds=bounds, **options
    )
    return result


def _pem_solve_bfgs(
    pem_obj: CompoundPEMObjective,
    p_guess: T,
    bounds: tuple[T, T] | None = None,
    options: dict | None = None,
) -> OptimizeResult:
    method = "BFGS" if bounds is None else "L-BFGS-B"

    if options is None:
        options = {}

    p_guess_flat, unravel = arc.tree.ravel(p_guess)

    if bounds is not None:
        lb, ub = bounds
        lb_flat, _ = arc.tree.ravel(lb)
        ub_flat, _ = arc.tree.ravel(ub)
        # Zip bounds into (lb, ub) for each parameter
        bounds = list(zip(lb_flat, ub_flat))  # type: ignore

    # Define an objective and gradient function for BFGS
    @arc.compile
    def func(params_flat):
        return pem_obj(unravel(params_flat))

    jac = arc.grad(func)

    # Set sensible defaults for system identification problems
    default_options = {
        "gtol": 1e-6,
        "disp": True,
        "maxiter": 200,
    }

    options = {**default_options, **options}

    result: OptimizeResult = scipy_minimize(
        func,
        p_guess_flat,
        method=method,
        jac=jac,
        bounds=bounds,
        options=options,
    )

    # Replace the flat parameters with the original tree structure
    result.x = unravel(result.x)
    return result


def _pem_solve_ipopt(
    pem_obj: CompoundPEMObjective,
    p_guess: T,
    bounds: tuple[T, T] | None = None,
    options: dict | None = None,
) -> OptimizeResult:
    if options is None:
        options = {}

    # Set sensible defaults for system identification problems
    ipopt_default_options = {
        "tol": 1e-6,
        "max_iter": 200,
        "hessian_approximation": "limited-memory",
    }
    default_options: dict[str, str | float | int] = {}

    options["ipopt"] = {**ipopt_default_options, **options.get("ipopt", {})}
    options = {**default_options, **options}

    p_guess_flat, unravel = arc.tree.ravel(p_guess)
    if bounds is not None:
        lb, ub = bounds
        lb_flat, _ = arc.tree.ravel(lb)
        ub_flat, _ = arc.tree.ravel(ub)
        bounds = (lb_flat, ub_flat)  # type: ignore

    # Define an objective and gradient function for BFGS
    @arc.compile
    def func(params_flat):
        return pem_obj(unravel(params_flat))

    result: OptimizeResult = arc.minimize(
        func,
        p_guess_flat,
        bounds=bounds,
        options=options,
    )

    result.x = unravel(result.x)
    result.nit = -1  # Placeholder for number of iterations
    result.fun = pem_obj(result.x)

    return result


SUPPORTED_METHODS = {
    "lm": _pem_solve_lm,
    "bfgs": _pem_solve_bfgs,
    "ipopt": _pem_solve_ipopt,
}


def pem(
    predictor: KalmanFilterBase,
    data: Timeseries | tuple[Timeseries, ...],
    p_guess: T,
    x0: np.ndarray | tuple[np.ndarray, ...],
    estimate_x0: bool | tuple[bool, ...] = False,
    bounds: tuple[T, T] | None = None,
    P0: np.ndarray | None | tuple[np.ndarray | None, ...] = None,
    method: str = "lm",
    options: dict | None = None,
) -> OptimizeResult:
    """Estimate parameters using Prediction Error Minimization.

    Solves the system identification problem by minimizing the prediction
    error between model predictions and measured outputs using a Kalman
    filter framework. This approach provides optimal handling of process
    and measurement noise while enabling efficient gradient computation
    through automatic differentiation.

    The method implements the discrete-time prediction error objective:

    .. code-block:: text

        minimize  J = (1/N) Σ[k=1 to N] e[k]ᵀ e[k]

    where ``e[k] = y[k] - ŷ[k|k-1]`` are the one-step-ahead prediction errors
    (innovations) from the Kalman filter and ``N`` is the number of measurements.

    This formulation automatically accounts for:

    - **Noise handling**: Process and measurement noise are modeled explicitly
    - **Recursive estimation**: Kalman filter provides efficient state propagation
    - **Gradient computation**: Automatic differentiation through filter recursions

    Parameters
    ----------
    predictor : KalmanFilterBase
        Kalman filter implementing the system model. Must provide
        ``step(t, x, y, P, args)`` method. Common choices:

        - :class:`ExtendedKalmanFilter`
        - :class:`UnscentedKalmanFilter`

        The predictor encapsulates the system dynamics, observation model,
        and noise characteristics (``Q``, ``R`` matrices).
    data : Timeseries
        Input-output data containing synchronized time series:

        - ``ts`` : Time vector of shape ``(N,)``
        - ``us`` : Input signals of shape ``(nu, N)``
        - ``ys`` : Output measurements of shape ``(ny, N)``

        All arrays must have consistent time dimensions.  Multiple
        :class:`Timeseries` instances can be provided as a tuple for
        multi-experiment identification, allowing joint parameter estimation
        across different datasets.

        If multiple datasets are provided, the initial state estimates
        (``x0``) should also be a tuple of initial conditions corresponding
        to each dataset.
    p_guess : Tree
        Initial parameter guess with arbitrary nested structure
        (e.g., ``{"mass": 1.0, "damping": {"c1": 0.1, "c2": 0.2}}``).
        The optimization preserves this structure in the result, enabling
        natural organization of physical parameters.
    x0 : array_like, optional
        Initial state estimate of shape ``(nx,)``.  Used as an initial guess
        if ``estimate_x0=True``.  For multiple datasets, this can be a tuple
        of initial conditions, allowing different initial states for each
        dataset.
    estimate_x0 : bool, default=False
        Whether to estimate the initial state ``x0`` along with parameters.
    bounds : tuple of (Tree, Tree), optional
        Parameter bounds as ``(lower_bounds, upper_bounds)`` with the
        same tree structure as ``p_guess``. Enables physical
        constraints such as:

        - Positive masses, stiffnesses, damping coefficients
        - Bounded gain parameters, time constants
        - Realistic physical parameter ranges

        Use ``-np.inf`` and ``np.inf`` for unbounded parameters.
    P0 : array_like, optional
        Initial state covariance matrix of shape ``(nx, nx)``. If None,
        defaults to identity matrix. Represents uncertainty in initial
        state estimate.  Can be a tuple of matrices if multiple
        datasets are provided, allowing different initial uncertainties
        for each dataset.
    method : str, default="bfgs"
        Optimization method. Currently only "lm" (Levenberg-Marquardt), "ipopt",
        and "bfgs" (BFGS) are supported. The "lm" method is a custom implementation
        roughly based on the MINPACK algorithm, and the "bfgs" method dispatches to
        a SciPy wrapper ("BFGS" or "L-BFGS-B", depending on whether there are bounds).
    options : dict, optional
        Optimization options passed to the underlying optimization solver.

        For the "lm" method, these include:

        - ``ftol`` : Function tolerance (default: 1e-4)
        - ``xtol`` : Parameter tolerance (default: 1e-6)
        - ``gtol`` : Gradient tolerance (default: 1e-6)
        - ``max_nfev`` : Maximum function evaluations (default: 200)
        - ``nprint`` : Progress printing interval (default: 0)

        For the "bfgs" method, these include:

        - ``gtol`` : Gradient tolerance (default: 1e-6)
        - ``disp`` : Whether to print convergence information (default: True)
        - ``maxiter`` : Maximum iterations (default: 200)
        - ``hess_inv0`` : Initial Hessian inverse for BFGS (optional)

        If ``hess_inv0`` is not provided, a Gauss-Newton-like Hessian approximation
        is used to initialize the BFGS inverse-Hessian approximation.

        For the "ipopt" method, see the :func:`archimedes.minimize` documentation
        for available options. The solver defaults to a limited-memory approximation
        of the Hessian.

    Returns
    -------
    result : scipy.optimize.OptimizeResult
        Optimization result with estimated parameters in ``result.x``
        preserving the original tree structure. Additional fields include:

        - ``success`` : Whether estimation succeeded
        - ``fun`` : Final prediction error objective value
        - ``nit`` : Number of optimization iterations
        - ``history`` : Detailed convergence history

    Notes
    -----
    This implementation provides:

    **Automatic Gradients**:
        Efficient gradient computation through automatic differentiation of
        the Kalman filter recursions. No need for finite differences or
        manual Jacobian implementation.

    **Structured Parameters**:
        Natural handling of nested parameter dictionaries enables intuitive
        organization of physical parameters and parameter bounds.

    **Physical Constraints**:
        Box constraints enable realistic parameter bounds (mass > 0, etc.)
        without sacrificing convergence properties.

    **Kalman Filter Integration**:
        Seamless integration with both Extended and Unscented Kalman Filters
        enables handling of linear and nonlinear systems with appropriate
        accuracy-efficiency tradeoffs.

    **Multi-Experiment Support**:
        Can handle multiple datasets simultaneously, allowing joint parameter
        estimation across different experiments or operating conditions.

    The method automatically computes gradients with respect to both initial
    conditions (when ``estimate_x0=True``) and model parameters using efficient
    automatic differentiation through the Kalman filter recursions. This
    avoids the computational expense and numerical issues of finite difference
    approximations.

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>> from archimedes.sysid import pem, Timeseries
    >>> from archimedes.observers import ExtendedKalmanFilter
    >>>
    >>> # Define second-order damped oscillator
    >>> def dynamics(t, x, u, params):
    ...     omega_n = params["omega_n"]  # Natural frequency
    ...     zeta = params["zeta"]        # Damping ratio
    ...
    ...     return np.hstack([
    ...         x[1],  # velocity
    ...         -omega_n**2 * x[0] - 2*zeta*omega_n*x[1] + omega_n**2 * u[0]
    ...     ])
    >>>
    >>> def observation(t, x, u, params):
    ...     return x[0]  # Measure position only
    >>>
    >>> # Generate synthetic measurement data
    >>> dt = 0.05
    >>> ts = np.arange(0, 10, dt)
    >>> x0 = np.array([0.0, 0.0])
    >>> params_true = {"omega_n": 2.0, "zeta": 0.1}
    >>> us = np.ones((1, len(ts)))  # Step input
    >>> # Generate step response
    >>> xs = arc.odeint(
    ...     dynamics,
    ...     (ts[0], ts[-1]),
    ...     x0,
    ...     t_eval=ts,
    ...     args=(np.array([1.0]), params_true),
    ... )
    >>> noise_std = 0.01
    >>> ys = xs[:1, :] + np.random.normal(0, noise_std, size=xs.shape[1])
    >>>
    >>> # Set up identification problem
    >>> dyn_discrete = arc.discretize(dynamics, dt, method="rk4")
    >>> Q = noise_std ** 2 * np.eye(2)  # Process noise covariance
    >>> R = noise_std ** 2 * np.eye(1)  # Measurement noise covariance
    >>> ekf = ExtendedKalmanFilter(dyn_discrete, observation, Q, R)
    >>>
    >>> data = Timeseries(ts=ts, us=us, ys=ys)
    >>> p_guess = {"omega_n": 2.5, "zeta": 0.5}
    >>>
    >>> # Estimate parameters with known initial conditions
    >>> result = pem(ekf, data, p_guess, x0=x0)
    >>> print(f"Estimated parameters: {result.p}")
    Estimated parameters: {'omega_n': array(1.9709515), 'zeta': array(0.11517324)}
    >>> print(f"Converged in {result.nit} iterations")
    Converged in 26 iterations
    >>>
    >>> # With physical parameter constraints
    >>> bounds = (
    ...     {"omega_n": 0.0, "zeta": 0.0},     # Lower bounds (positive values)
    ...     {"omega_n": 10.0, "zeta": 1.0},    # Upper bounds (reasonable ranges)
    ... )
    >>> result = pem(ekf, data, p_guess, x0=x0, bounds=bounds)

    See Also
    --------
    archimedes.optimize.lm_solve : Underlying Levenberg-Marquardt optimizer
    Timeseries : Data container for input-output time series
    archimedes.observers.ExtendedKalmanFilter : EKF for mildly nonlinear systems
    archimedes.observers.UnscentedKalmanFilter : UKF for highly nonlinear systems
    archimedes.discretize : Convert continuous-time dynamics to discrete-time

    References
    ----------
    .. [1] Ljung, L. "System Identification: Theory for the User." 2nd edition,
           Prentice Hall, 1999.
    """
    if method not in SUPPORTED_METHODS:
        raise ValueError(f"Unsupported method: {method}.")

    pem_solve = SUPPORTED_METHODS[method]

    if not isinstance(data, (tuple, list)):
        data = (data,)

    n_expt = len(data)

    # For each argument, replicate for all experiments if needed

    if isinstance(x0, np.ndarray):
        x0 = tuple([x0] * n_expt)

    if P0 is None or isinstance(P0, np.ndarray):
        P0 = tuple([P0] * n_expt)

    if isinstance(estimate_x0, bool):
        estimate_x0 = tuple([estimate_x0] * n_expt)

    sub_objectives = []
    x0_guess: list[np.ndarray | None] = []
    for i in range(n_expt):
        if estimate_x0[i]:
            x0_guess.append(x0[i])
            sub_x0 = None
        else:
            x0_guess.append(None)
            sub_x0 = x0[i]

        sub_objectives.append(
            PEMObjective(
                predictor=predictor,
                data=data[i],
                P0=P0[i],
                x0=sub_x0,
            )
        )

    objective = CompoundPEMObjective(sub_objectives=sub_objectives)
    dvs_guess = (x0_guess, p_guess)

    if bounds is not None:
        x0_bounds = [None] * n_expt
        lb, ub = bounds
        lb = (x0_bounds, lb)  # type: ignore
        ub = (x0_bounds, ub)  # type: ignore
        bounds = (lb, ub)

    result: OptimizeResult = pem_solve(
        pem_obj=objective,
        p_guess=dvs_guess,
        bounds=bounds,
        options=options,
    )

    x0_opt, params_opt = result.x
    delattr(result, "x")
    result.p = params_opt

    if n_expt == 1:
        x0_opt = x0_opt[0]

    result.x0 = x0_opt  # Store optimized initial condition

    return result



# ruff: noqa: N806, N803, N815
from __future__ import annotations

import numpy as np

from archimedes import struct

from ._attitude import Attitude

__all__ = [
    "RigidBody",
]


@struct
class RigidBody:
    """6-dof rigid body dynamics model

    This class implements 6-dof rigid body dynamics based on reference equations
    from Lewis, Johnson, and Stevens, "Aircraft Control and Simulation" [1]_.

    This implementation is general and does not make any assumptions about the
    forces, moments, or mass properties.  These must be provided as inputs to the
    dynamics function.

    The model does assume a non-inertial body-fixed reference frame B and a Newtonian
    inertial reference frame N.  The body frame is assumed to be located at the
    vehicle's center of mass.

    With these conventions, the state vector is defined as
        ``x = [pos, att, v_B, w_B]``

    where

    - ``pos`` = position of the center of mass in the inertial frame
    - ``att`` = attitude (orientation) of the vehicle
    - ``v_B`` = velocity of the center of mass in body frame B
    - ``w_B`` = angular velocity in body frame (ω_B)

    Note that the attitude can be any object implementing the :py:class:`Attitude`
    protocol, commonly :py:class:`Quaternion` or :py:class:`EulerAngles`.
    The transformation implemented by ``as_matrix`` with this attitude represents
    ``R_BN``, the rotation from the inertial frame N to the body frame B.

    The equations of motion for a quaternion attitude are given by

    .. math::
        \\dot{\\mathbf{p}}^N &= \\mathbf{R}_{BN}^T(\\mathbf{q}) \\mathbf{v}^B \\\\
        \\dot{\\mathbf{q}} &= \\frac{1}{2} \\mathbf{q} \\otimes \\boldsymbol{\\omega}^B
            \\\\
        \\dot{\\mathbf{v}}^B &= \\frac{1}{m}\\mathbf{F}^B
            - \\boldsymbol{\\omega}^B \\times \\mathbf{v}^B \\\\
        \\dot{\\boldsymbol{\\omega}}^B &= \\mathbf{J}_B^{-1}(\\mathbf{M}^B
            - \\boldsymbol{\\omega}^B \\times (\\mathbf{J}^B \\boldsymbol{\\omega}^B))

    where

    - :math:`\\mathbf{R}_{BN}(\\mathbf{q})` = direction cosine matrix (DCM)
    - :math:`m` = mass of the vehicle
    - :math:`\\mathbf{J}^B` = inertia matrix of the vehicle in body axes
    - :math:`\\mathbf{F}^B` = net forces acting on the vehicle in body frame B
    - :math:`\\mathbf{M}^B` = net moments acting on the vehicle in body frame B

    The inputs to the dynamics function are a ``RigidBody.Input`` struct
    containing the forces, moments, mass, and inertia properties.  By default
    the time derivatives of the mass and inertia are zero unless specified
    in the input struct.

    The attitude representation can be anything that implements the
    :py:class:`Attitude` protocol, which requires methods for rotating vectors
    between frames and calculating the attitude kinematics.  Common choices
    are :py:class:`Quaternion` or :py:class:`EulerAngles`.
    The time derivative of the attitude will be calculated with the classes'
    own kinematics method, so for example Euler angle rates will be subject to
    the usual gimbal lock singularity.

    Examples
    --------
    
    This class is implemented as a "singleton", meaning it does not need to
    actually be instantiated and all methods are class methods.

    >>> import archimedes as arc
    >>> from archimedes.spatial import RigidBody, Quaternion
    >>> import numpy as np
    >>> t = 0
    >>> v_B = np.array([1, 0, 0])  # Constant velocity in x-direction
    >>> att = Quaternion([1, 0, 0, 0])  # No rotation
    >>> x = RigidBody.State(
    ...     pos=np.zeros(3),
    ...     att=att,
    ...     v_B=v_B,
    ...     w_B=np.zeros(3),
    ... )
    >>> u = RigidBody.Input(
    ...     F_B=np.array([0, 0, -9.81]),  # Gravity
    ...     M_B=np.zeros(3),
    ...     m=2.0,
    ...     J_B=np.diag([1.0, 1.0, 1.0]),
    ... )
    >>> RigidBody.dynamics(t, x, u)
    State(pos=array([1., 0., 0.]),
      att=Quaternion([0., 0., 0., 0.]),
      v_B=array([ 0.   ,  0.   , -4.905]),
      w_B=array([0., 0., 0.]))

    A common pattern is to have a vehicle model inherit from ``RigidBody.State``,
    even while the model class itself does not inherit from ``RigidBody``.  This
    allows the vehicle model to have its own state representation while still being
    compatible with the rigid body dynamics:

    >>> @arc.struct
    ... class AircraftState(RigidBody.State):
    ...     eng: np.ndarray  # engine state
    ...
    >>> state = AircraftState(
    ...     pos=np.zeros(3),
    ...     att=Quaternion([1, 0, 0, 0]),
    ...     v_B=np.zeros(3),
    ...     w_B=np.zeros(3),
    ...     eng=np.array([0.0]),
    ... )
    >>> x_dot_rb = RigidBody.dynamics(t, state, u)
    >>> x_dot_rb
    State(pos=array([0., 0., 0.]),
      att=Quaternion([0., 0., 0., 0.]),
      v_B=array([ 0.   ,  0.   , -4.905]),
      w_B=array([0., 0., 0.]))

    The ``dynamics`` method returns a ``RigidBody.State`` struct containing only the
    derivatives of rigid body states.  To use this in a vehicle model with additional
    states, you can manually combine the rigid body state derivatives with the
    derivatives of the additional states.

    >>> x_dot = AircraftState(
    ...     pos=x_dot_rb.pos,
    ...     att=x_dot_rb.att,
    ...     v_B=x_dot_rb.v_B,
    ...     w_B=x_dot_rb.w_B,
    ...     eng=np.array([0.1]),  # engine state derivative
    ... )
    >>> x_dot
    AircraftState(pos=array([0., 0., 0.]),
      att=Quaternion([0., 0., 0., 0.]),
      v_B=array([ 0.   ,  0.   , -4.905]),
      w_B=array([0., 0., 0.]),
      eng=array([0.1]))

    Notes
    -----

    The equations of motion implemented here are technically correct only for the
    case of a rigid body with constant mass, inertia, and center of gravity moving
    in an inertial reference frame and without "internal" angular velocity (gyroscopic
    effects).  However, the model can be extended to account for these effects if
    needed by passing pseudo-forces and moments.

    In all the following cases, the effects can be treated as constant, quasi-steady
    (time-varying but with negligible rates), or fully dynamic (time-varying with
    non-negligible rates).  In both cases, the current value and time derivatve should
    be tracked and computed outside of the rigid body model, and the appropriate
    values passed in the input struct.

    - **Variable mass**: Quasi-steady mass may be handled by passing the current mass
        in the input struct.  The mass rate of change :math:`\\dot{m}` enters the
        equations of motion via the time derivative of linear momentum:

        .. math::
            \\frac{d}{dt}(m \\mathbf{v}^B) = \\mathbf{F}^B
            \\implies m \\dot{\\mathbf{v}}^B + \\dot{m} \\mathbf{v}^B = \\mathbf{F}^B

        Hence, mass flow rates can be accounted for by including the pseudo-force
        :math:`-\\dot{m} \\mathbf{v}^B` in the net forces passed as input.

    - **Variable inertia**: In the same way, quasi-steady inertia may be handled
        by passing the current inertia matrix in the input struct.  The inertia rate
        of change :math:`\\dot{\\mathbf{J}}^B` enters the equations of motion via the
        time derivative of angular momentum:

        .. math::
            \\frac{d}{dt}(\\mathbf{J}^B \\boldsymbol{\\omega}^B) = \\mathbf{M}^B
            \\implies \\mathbf{J}^B \\dot{\\boldsymbol{\\omega}}^B
            + \\dot{\\mathbf{J}}^B \\boldsymbol{\\omega}^B = \\mathbf{M}^B

        Non-negligible inertia rates can be accounted for by including the
        pseudo-moment :math:`-\\dot{\\mathbf{J}}^B \\boldsymbol{\\omega}^B` in the
        net moment passed as input.

    - **Variable center of mass**: The equations of motion are derived about the
        center of mass (CM).  However, typically the body-fixed reference frame B is
        defined at some convenient reference point that may not coincide with the
        instantaneous center of mass.  Properties like aerodynamics and propulsion
        behaviors are also often defined with respect to the reference CM.

        If the reference CM is at the origin of the body frame B and the actual CM
        is at a point :math:`\\mathbf{r}_{CM}^B` in body frame B moving with velocity
        :math:`\\dot{\\mathbf{r}}_{CM}^B` with respect to the reference point, then
        the relationship between the state velocity :math:`\\mathbf{v}^B` (that is, the
        inertial velocity of the CM expressed in body frame B) and the velocity of
        the reference point :math:`\\mathbf{v}_{ref}^B` is

        .. math::
            \\mathbf{v}^B = \\mathbf{v}_{ref}^B +
            \\dot{\\mathbf{r}}_{CM}^B + \\boldsymbol{\\omega}^B \\times
            \\mathbf{r}_{CM}^B

        Often this correction is negligible, but if needed then the state velocity
        should be converted to the reference point velocity before computing
        aerodynamics or other quantities referenced to the body frame origin.
        In the common case that the CM is moving due to fuel consumption or payload
        release, the relative velocity :math:`\\dot{\\mathbf{r}}_{CM}^B` is usually
        negligible.

        A more important effect is the moment transfer from the offset of the forces
        acting at the reference point to the actual CM.  If the net force acting on
        the vehicle at the reference point is :math:`\\mathbf{F}_{ref}^B`, then the
        moment about the CM is given by

        .. math::
            \\mathbf{M}^B = \\mathbf{M}_{ref}^B -
            \\mathbf{r}_{CM}^B \\times \\mathbf{F}_{ref}^B

        The same transformation applies to forces computed about an arbitrary reference
        point, but the moment arm will then be the vector from that reference point to
        the instantaneous CM.

    - **Gyroscopic effects**: The full Euler equation for rotational dynamics in a
        non-inertial body-fixed frame is

        .. math::
            \\mathbf{M}^B = \\frac{d\\mathbf{h}^B}{dt} + \\boldsymbol{\\omega}^B
            \\times \\mathbf{h}^B,

        where :math:`\\mathbf{h}^B` is the net angular momentum of the vehicle in the
        body frame B.  If the vehicle does not have any "internal" angular momentum,
        then :math:`\\mathbf{h}^B = \\mathbf{J}^B \\boldsymbol{\\omega}^B` and the
        equations reduce to those implemented here.
        
        However, if there are significant additional contributions to angular momentum,
        these affect the dynamics via gyroscopic pseudo-moments.  If a system has
        internal angular momentum :math:`\\mathbf{h}_{int}^B =
        \\sum_{i} \\mathbf{J}_{int,i}^B \\boldsymbol{\\omega}_{int,i}^B`, these
        contributions must be included:

        .. math::
            \\mathbf{M}^B = \\frac{d}{dt}(\\mathbf{J}^B \\boldsymbol{\\omega}^B)
            + \\frac{d\\mathbf{h}_{int}^B}{dt}
            + \\boldsymbol{\\omega}^B \\times \\mathbf{J}^B \\boldsymbol{\\omega}^B
            + \\boldsymbol{\\omega}^B \\times \\mathbf{h}_{int}^B

        The additional terms involving :math:`\\mathbf{h}_{int}^B` can be treated as
        pseudo-moments and included in the net moment passed as input.  The usual
        logical flow would be to compute both the internal angular momentum and its
        time derivative outside of the rigid body model (e.g. as a subsystem
        calculation), and then pass the net effective moment

        .. math::
            \\mathbf{M}_\\mathrm{eff}^B = \\mathbf{M}^B
            - \\frac{d}{dt}(\\mathbf{h}_{int}^B)
            - \\boldsymbol{\\omega}^B \\times \\mathbf{h}_{int}^B

        as the input to the rigid body dynamics.

        For example, a calculation of the gyroscopic effects of a spinning rotor
        with inertia :math:`\\mathbf{J}_\\mathrm{rot}^B`, angular velocity
        :math:`\\boldsymbol{\\omega}_\\mathrm{rot}^B`, and negligible angular
        acceleration might look like:

        .. code-block:: python

            h_int_B = J_rot_B @ w_rot_B  # Rotor angular momentum

            # Compute effective moment including gyroscopic effects
            M_eff_B = M_B - np.cross(w_B, h_int_B)

    - **Non-inertial reference frame**: These equations of motion are valid only
        when referenced to a Newtonian inertial frame N.  This is of course an
        idealization in all cases, but it is always possible to find _some_ frame
        that is nearly enough inertial for modeling purposes.

        However, a common situation in aerospace applications is to model a body
        moving relative to a rotating planetary frame E (e.g. the Earth-centered,
        Earth-fixed frame ECEF) that is assumed to be in non-accelerating but
        rotating with some angular velocity :math:`\\boldsymbol{\\Omega}_{E}` with
        respect to the inertial frame N.  In this case an alternative formulation
        uses a state vector composed of:

        - :math:`\\mathbf{p}^E` = position of the center of mass in the frame E
        - :math:`\\mathbf{q}` = attitude (orientation) of the vehicle with respect to E
        - :math:`\\mathbf{v}^E` = velocity of the center of mass in rotating frame E
        - :math:`\\boldsymbol{\\omega}^B` = angular velocity in body frame (ω_B) with
            respect to the inertial frame N

        Then the equations of motion are [1]_:

        .. math::
            \\dot{\\mathbf{p}}^E &= \\mathbf{v}^E \\\\
            \\dot{\\mathbf{q}} &= \\frac{1}{2} \\mathbf{q} \\otimes \\left(
                \\boldsymbol{\\omega}^B - \\boldsymbol{\\Omega}_{E}^B \\right) \\\\
            \\dot{\\mathbf{v}}^E &= \\frac{1}{m}\\mathbf{F}^E
                - 2 \\boldsymbol{\\Omega}_{E}^E \\times \\mathbf{v}^E
                - \\boldsymbol{\\Omega}_{E}^E \\times
                (\\boldsymbol{\\Omega}_{E}^E \\times \\mathbf{p}^E)
            \\dot{\\boldsymbol{\\omega}}^B &= \\mathbf{J}_B^{-1}(\\mathbf{M}^B
                - \\boldsymbol{\\omega}^B \\times (\\mathbf{J}^B
                \\boldsymbol{\\omega}^B))

        Unfortunately, this cannot be straightforwardly reconciled with the
        implementation here, even with the addition of the Coriolis and centrifugal
        pseudo-forces.  This is because of the definition of the attitude and angular
        velocity with respect to different reference frames (E and N, respectively).
        Using the angular velocity relative to frame N allows the use of the Euler
        dynamics equation without complex pseudo-moments, but means that the angular
        velocity must be modified by :math:`-\\boldsymbol{\\Omega}_{E}^B` in the
        attitude kinematics.

        The equations above could be implemented in an ECEF frame with a simple custom
        rigid body class:

        .. code-block python::

            @struct
            class EarthReferencedBody:
                rot_earth: float = 7.292e-5

                @struct
                class State:
                    pos: np.ndarray  # ECEF position
                    att: Attitude  # Body attitude relative to ECEF
                    v_E: np.ndarray  # ECEF velocity
                    w_B: np.ndarray

                @struct
                class Input:
                    F_E: np.ndarray  # Forces in Earth frame
                    M_B: np.ndarray  # Moments in body frame
                    m: float  # Mass
                    J_B: np.ndarray  # Inertia matrix in body frame

                def dynamics(self, t: float, x: State, u: Input) -> State:
                    Ω_E = np.hstack([0.0, 0.0, self.rot_earth])
                    R_BE = x.att.as_matrix()
                    v_E, p_E = x.v_E, x.pos  # ECEF position, velocity

                    # Position and attitude kinematics
                    pos_deriv = v_E
                    att_deriv = x.att.kinematics(x.w_B - R_BE @ Ω_E)

                    # Force equation with Coriolis and centrifugal effects
                    dv_E = (u.F_E / u.m) - np.cross(Ω_E, 2 * v_E - np.cross(Ω_E, p_E))
                    
                    # Moment equation (same as body-referenced formulation)
                    dw_B = np.linalg.solve(
                        u.J_B, u.M_B - np.cross(x.w_B, u.J_B @ x.w_B)
                    )

                    # Output time derivatives of substates
                    return RigidBody.State(
                        pos=pos_deriv, att=att_deriv, v_E=dv_E, w_B=dw_B
                    )

        While this formulation does cover a substantial number of orbital mechanics
        applications, it is not one-size-fits all.  Are centrifugal effects accounted
        for in the gravity model?  Are precession and nutation important?  Is the
        angular velocity time-varying?  The present design prioritizes _customization_
        over _comprehensiveness_.

        In short, handling of non-inertial frames in Archimedes still needs some
        design work and is not robustly supported.  The recommendation is to
        implement custom rigid body dynamics based on the above equations.  If you
        would like to see support for non-inertial frames be a higher priority,
        please feel free to raise the issue in the `Discussions
        <https://github.com/PineTreeLabs/archimedes/discussions>`__ page on GitHub.

    As a combined example of pseudo-force handling, if you have a flight dynamics
    model with constant inertia and quasi-steady mass and CM location, a typical
    calculation of the rigid body inputs might look like the following:

    .. code-block:: python

        @struct
        class FlightVehicle:
            dry_mass: float
            wet_mass: float
            J_B: np.ndarray  # constant inertia matrix
            dCM_dm: np.ndarray  # CM location rate w.r.t. mass [m/kg]
    
            @struct
            class State(RigidBody.State):
                fuel_mass: float
                ...

            @struct
            class Input:
                throttle: float
                ...

            def dynamics(t, x: State, u: Input) -> State:
                # Compute current mass and CM location
                m = self.dry_mass + x.fuel_mass
                r_CM_B = self.dCM_dm * (self.wet_mass - m)  # ref point -> CM

                # Velocity at reference point
                v_ref_B = x.v_B - np.cross(x.w_B, r_CM_B)

                # Aerodynamic forces and moments at body frame origin
                F_aero_B, M_aero_B = self.aerodynamics(v_ref_B, ...)

                # Propulsion forces and moments at body frame origin
                F_prop_B, M_prop_B = self.propulsion(v_ref_B, ...)

                # Net forces
                F_B = F_aero_B + F_prop_B

                # Net moments, including moment transfer of forces to CM
                M_B = M_aero_B + M_prop_B - np.cross(r_CM_B, F_B)

                # Compute rigid body dynamics
                rb_input = RigidBody.Input(F_B=F_B, M_B=M_B, m=m, J_B=self.J_B)
                x_dot_rb = RigidBody.dynamics(t, x, rb_input)

                # Combine with other state derivatives (e.g. fuel mass)
                return Vehicle.State(
                    pos=x_dot_rb.pos,
                    att=x_dot_rb.att,
                    v_B=x_dot_rb.v_B,
                    w_B=x_dot_rb.w_B,
                    fuel_mass=-self.propulsion.burn_rate(u),
                    ...
                )


    References
    ----------
    .. [1] Lewis, F. L., Johnson, E. N., & Stevens, B. L. (2015).
            Aircraft Control and Simulation. Wiley.
    """

    @struct
    class State:
        pos: np.ndarray  # Position of the center of mass in the inertial frame
        att: Attitude  # Attitude (orientation) of the vehicle
        v_B: np.ndarray  # Velocity of the center of mass in body frame B
        w_B: np.ndarray  # Angular velocity in body frame (ω_B)

    @struct
    class Input:
        F_B: np.ndarray  # Forces in body frame
        M_B: np.ndarray  # Moments in body frame
        m: float  # Mass
        J_B: np.ndarray  # Inertia matrix in body frame

    @classmethod
    def dynamics(cls, t: float, x: State, u: Input) -> State:
        """Calculate 6-dof dynamics

        Args:
            t: time
            x: dynamic state
            u: input struct with net forces and moments

        Returns:
            x_t: time derivative of the state vector
        """
        R_BN = x.att.as_matrix()
        pos_deriv = R_BN.T @ x.v_B
        att_deriv = x.att.kinematics(x.w_B)
        dv_B = (u.F_B / u.m) - np.cross(x.w_B, x.v_B)
        dw_B = np.linalg.solve(u.J_B, u.M_B - np.cross(x.w_B, u.J_B @ x.w_B))
        return RigidBody.State(pos=pos_deriv, att=att_deriv, v_B=dv_B, w_B=dw_B)
